/* This routine uses tables with observed profiles  */ 

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "structures.h"
#include "mem_alloc.h"
#include "constants.h"
#include "ubvri_40.h"
#include "spline.h"

#define defC 0.999

static float Cq0;  /* default value for Cq */ 
static float Cq1;  /* default value for Cq in spots (when reconstructing ff's) */ 
static int *nspec, *nspec_s; 
static float **spec, **spec_s, **lam, **lam_s, **y2, **y2_s; 
static void sample(float *I, double *lambda, int n, float Cq, int id); 
static void convert(float Cq, float *bright, float *dbright, int id); 
int locate(float x, float array[], int n); 
static void spl(double *x, float *y, int npt, float *lam, float *data, float *yy, int nspec); 
void linear(double *x, float *y, int npt, float *lam, float *data, int nspec); 
static int read_table(char filename[], int batch); 


/*  We use static variables to hold the line information.  */

#define NUM_LINES 9
static char *line_names[NUM_LINES] = { 
                                       "Mean",
                                       "Ca1_644",
                                       "Fe1_666",
                                       "Ca1_672",
                                       "Halpha",
		                       "B_band",
		                       "V_band",
		                       "R_band",
		                       "I_band" };
static char *empty = ""; 

static int line_ids[NUM_LINES] = {0,1,2,3,4,5,6,7,8};

static int phot[NUM_LINES] = {0,0,0,0,0,2,3,4,5};  
    /* 0 is for a spectral line, 1,2,3,4,5 are for UBVRcIc bands */ 

static int mod[NUM_LINES] = {0,0,0,0,0};  
    /* mod is modified when reading kurucz tables */ 

static int lin[NUM_LINES] = {1,0,0,0,0,1,1,1,1};

static int mag_opt[NUM_LINES] = {1,1,1,1,1}; 

static double l0[NUM_LINES]  = {          /* $\lambda_0$ */
                                 550.0000,
                                 643.9075,
                                 666.3330,
                                 671.7681,
                                 656.2797,
			 	 440.0000,
			 	 550.0000,
			 	 700.0000,
			 	 900.0000 }; 
static double eps[NUM_LINES] = {         /* $\epsilon$ */
                                 0.75,0.75,0.75,0.75,0.60,0.85,0.80,0.70,0.60};
static double g[NUM_LINES] =   {         /* Lande g factor */
                                 0.00,0.00,0.00,0.00,0.00}; 
static double zeta_rt[NUM_LINES] = {     /* $\zeta$ */
                                 0.00,0.00,0.00,0.00,0.00}; 
static float bright_phot[NUM_LINES], bright_spot[NUM_LINES]; 


void init_spectral_lines(batch)
int batch;
{
    int   err;  
    char filename[80]; 

    do {
        err = read_table(filename, batch); 
        if (err)
            if (batch == 1) {
                if (err == 1) printf("File not found.\n");
                else          printf("Wrong file.\n");
                exit(1);
            } else { 
                if (err == 1) printf("File not found.  Try again\n");
                else          printf("Wrong file.  Try again\n");
            } 
    } while (err);
}


void end_spectral_lines()
{
    int j; 

    for (j = 0; j < NUM_LINES; j++) {
        if (nspec[j]) { 
            free_vector(lam[j],  0, nspec[j]-1); 
            free_vector(spec[j], 0, nspec[j]-1); 
            free_vector(y2[j], 0, nspec[j]-1);
        }
        if (nspec_s[j]) { 
            free_vector(lam_s[j],  0, nspec_s[j]-1); 
            free_vector(spec_s[j], 0, nspec_s[j]-1); 
            free_vector(y2_s[j],   0, nspec_s[j]-1); 
        } 
    } 
    free_svector((void *) lam); 
    free_svector((void *) lam_s); 
    free_svector((void *) spec); 
    free_svector((void *) spec_s); 
    free_svector((void *) y2); 
    free_svector((void *) y2_s); 
    free_ivector(nspec, 0, NUM_LINES-1); 
    free_ivector(nspec_s, 0, NUM_LINES-1); 
}


void get_spectral_lines(nlines,names,ids)
int *nlines,**ids;
char **names[];
{
    *nlines = NUM_LINES;
    *names = line_names;
    *ids   = line_ids;
}


/*
The following routine lets the main program find out some details about
each of the lines.
*/

void line_info(id,ll0,dll,broadened,linear)
int id;      /* Line ID number */
double *ll0;  /* Line centre wavelength */
double *dll;  /* Doppler width */
int *broadened, *linear;  /* Does the line model include the instrumental broadening */
{
    *ll0 = l0[id];
    *dll = 0.0;     /* Doppler width dependant on Cq */ 
    *broadened = 1;
    *linear = lin[id]; 
}


/*
The function line_centre() needs to return the centre wavelength of the
line being studied.
*/

double find_line_centre(id)
int id;
{
    return l0[id];
}


/*
The function find_macro_turb() needs to return the radial-tangential
macroturbulence parameter.  If you don't need to use macroturbulence
just return zero.  If you do need it you should look at the routines
doppler_spec() and doppler_deriv() to check how it is being
implimented.
*/

double find_macro_turb(id)
int id;
{
    return zeta_rt[id];
}


/*
This routine is used when a spectrum has to be calculated from an image.
You only need to calculate the Stokes parameters that you will use.  For
example, in straight Doppler imaging you only have to calculate Stokes I
using the continuum brightness C_q.
*/

void line_profile(id,I,Q,U,V,n,lambda,Cq,Cm,mu,B,gam,chi,magnetic)
int id;                /* Line ID number */
float I[],Q[],U[],V[];  /* Returned Stokes parameters */
int n;                /* Number of wavelengths to calculate */
double lambda[];        /* Wavelength $\lambda$ */
float Cq,Cm;          /* Continuum brightness */
float mu;              /* Cosine of the limb angle */
float B,gam,chi;     /* Magnetic field strength and orientation $(B,\gamma,\chi)$ */
int magnetic;          /* Switch is non-zero for a magnetic star */
{
    int    j;
    float *dI, *tmp; 
    float bright,dbright,bright_s,dbright_s,contin;
/*
    convert(Cq0, &bright, &dbright, id); 
    convert(Cq1, &bright_s, &dbright_s, id); 
*/
bright = bright_phot[id]; bright_s = bright_spot[id]; 

    if (!phot[id]) { /* Spectral line */
        if (magnetic) {
            printf("This is not the correct line profile routine to use\n"); 
            exit(1); 
        } else { 
            contin = (1.0-eps[id]+eps[id]*mu)/(1.0-eps[id]/3.0);
            tmp = vector(0, n-1); 
            sample(I, lambda, n, Cq0, id); 
            sample(tmp, lambda, n, Cq1, id); 
            for (j = 0; j < n; j++) 
                I[j] = (Cq*bright*I[j]+(defC-Cq)*bright_s*tmp[j])*contin/defC; 
            free_vector(tmp, 0, n-1); 
        }

    } else { /* Photometry */
        contin = (1.0-eps[id]+eps[id]*mu)/(1.0-eps[id]/3.0);
        I[0] = (Cq*bright+(defC-Cq)*bright_s)*contin/defC; 
    }
}


/*
This routine is an extended version of line_profile() used during the
image inversion process.  As well as returning the spectral profile you
also need to return the derivatives of the field profile with respect
to the image quantities you are fitting.
*/

void line_deriv(id,dII,dQQ,dUU,dVV,I,Q,U,V,n,lambda,Cq,Cm,mu,B,gam,chi,magnetic)
int id;
float **dII,**dQQ,**dUU,**dVV;  /* Arrays of profile derivatives */
float I[],Q[],U[],V[];  /* Returned Stokes parameters */
int   n;          /* Number of wavelength points */
double lambda[];    /* Wavelength $\lambda$ */
float Cq,Cm;        /* Continuum brightness */
float mu;           /* Cosine of the limb angle */
float B,gam,chi;    /* Magnetic field strength and orientation $(B,\gamma,\chi)$ */
int magnetic;       /* Switch is non-zero for a magnetic star */
{
    int    j;
    float *dI,*tmp; 
    float bright,dbright,bright_s,dbright_s,contin;
/*
    convert(Cq0, &bright, &dbright, id); 
    convert(Cq1, &bright_s, &dbright_s, id); 
*/
bright = bright_phot[id]; bright_s = bright_spot[id]; 

    if (!phot[id]) { /* Spectral line */
        if (magnetic) {
            printf("This is not the correct line profile routine to use\n"); 
            exit(1); 
        } else { 
            contin = (1.0-eps[id]+eps[id]*mu)/(1.0-eps[id]/3.0);
            tmp = vector(0, n-1); 
            sample(I, lambda, n, Cq0, id); 
            sample(tmp, lambda, n, Cq1, id); 
            for (j = 0; j < n; j++) {
                dII[0][j] = (bright*I[j]-bright_s*tmp[j])*contin/defC; 
                I[j] = (Cq*bright*I[j]+(defC-Cq)*bright_s*tmp[j])*contin/defC; 
            }
            free_vector(tmp, 0, n-1); 
        }

    } else { /* Photometry */
        contin = (1.0-eps[id]+eps[id]*mu)/(1.0-eps[id]/3.0);
        I[0] = (Cq*bright+(defC-Cq)*bright_s)*contin/defC; 
        dII[0][0] = (bright-bright_s)*contin/defC; 
    }
}


void continuum(id,mu,Cq,Cm,cont,dCq,dCm,magnetic)
int id,magnetic;
float mu,Cq,Cm,*cont,*dCq,*dCm;
{
    float bright, dbright, bright_s, dbright_s, limb; 

    limb  = (1.0-eps[id]+eps[id]*mu)/(1.0-eps[id]/3.0);
/*
    convert(Cq0, &bright, &dbright, id); 
    convert(Cq1, &bright_s, &dbright_s, id); 
*/
bright = bright_phot[id]; bright_s = bright_spot[id]; 

    *cont = (Cq*bright+(defC-Cq)*bright_s)*limb/defC; 
    *dCq  = (bright-bright_s)*limb/defC; 
}


void image_info(Cm_used,qnames,ff_entropy,ff_rec,batch)
int *Cm_used,*ff_entropy,*ff_rec,batch;
char *qnames[];
{
    static char *fill = "filling factor";

    *Cm_used = 0; 
    *ff_entropy = *ff_rec = 1; 
    qnames[0] = fill;
}


void image_defaults(dCq, dB, magnetic, Cm_used, qnames, batch) 
float *dCq, *dB; 
int magnetic, Cm_used, batch; 
char *qnames[]; 
{
float junk; 
int j; 

    printf("Input default %s : ",qnames[0]);
    scanf("%f", dCq);
    if (batch) printf("%f\n",*dCq);
    printf("Input default temperature in photosphere : ", temp);
    scanf("%f", &Cq0);
    if (batch) printf("%f\n",Cq0);
    printf("Input default temperature in spots : ", temp);
    scanf("%f", &Cq1);
    if (batch) printf("%f\n",Cq1);

    for (j = 0; j < NUM_LINES; j++) {
        convert(Cq0, bright_phot+j, &junk, line_ids[j]); 
        convert(Cq1, bright_spot+j, &junk, line_ids[j]); 
        printf("%d %f %f\n", j, bright_phot[j], bright_spot[j]); 
    } 
} 


void transform(Cq,Cm,ngrid,magnetic)
float Cq[],Cm[];
int ngrid,magnetic;
{
}


void untransform(Cq,Cm,ngrid,magnetic)
float Cq[],Cm[];
int ngrid,magnetic;
{
}


void transform_photometry(spec_info, nobs, I, mag0, wphot)
spec_info_t *spec_info; 
int nobs; 
float *I, mag0, wphot; 
{
    int j, n; 

    for (j = n = 0; j < nobs; n += spec_info[j].nspec, j++) 
        if (spec_info[j].photometric && spec_info[j].selected) {
            I[n] = pow(10.0, -0.4*(I[n]-mag0)); 
            spec_info[j].SN = I[n]*(pow(10.0, 0.4*spec_info[j].SN)-1.0)/sqrt(wphot); 
        }
}


void untransform_photometry(spec_info, nobs, I, mag0, wphot)
spec_info_t *spec_info; 
int nobs; 
float *I, mag0, wphot; 
{
    int j, n; 

    for (j = n = 0; j < nobs; n += spec_info[j].nspec, j++) 
        if (spec_info[j].photometric && spec_info[j].selected) {
            spec_info[j].SN = 2.5*log10(1.0+spec_info[j].SN*sqrt(wphot)/I[n]); 
            I[n] = mag0-2.5*log10(I[n]); 
        }
}


static void convert(float Cq, float *bright, float *dbright, int id)
{
    float k, u, mag0, shift, fac; 
    int    j, j0; 

    if (!phot[id]) { /* Spectral lines */ 
        k = exp(1.439e7/(l0[id]*Cq0))-1.0; 
        u = exp(1.439e7/(l0[id]*Cq))-1.0; 
        *bright  = k/u; 
        *dbright = k*1.439e7*(u+1.0)/(l0[id]*u*u*Cq*Cq); 
   } else {      /* Photometry */ 
        shift = 0.0; /* compensate for errors in the calibration table */ 
        j0 = locate(Cq0+shift, temp, nt); 
        j  = locate(Cq+shift,  temp, nt); 
        mag0 = vv[j0] + (Cq0+shift-temp[j0])*(vv[j0+1]-vv[j0])/(temp[j0+1]-temp[j0]); 
        if (phot[id] == 2) {   /* B photometric band */ 
            *bright   = bb[j] + (Cq+shift-temp[j])*(bb[j+1]-bb[j])/(temp[j+1]-temp[j]); 
            *bright   = pow(10.0, -0.4 * (*bright-mag0));  
            *dbright  = db[j] + (Cq+shift-temp[j])*(db[j+1]-db[j])/(temp[j+1]-temp[j]); 
            *dbright *= -0.4 * log(10.0) * *bright; 
        } else if (phot[id] == 3) {   /* V photometric band */ 
            *bright   = vv[j] + (Cq+shift-temp[j])*(vv[j+1]-vv[j])/(temp[j+1]-temp[j]); 
            *bright   = pow(10.0, -0.4 * (*bright-mag0));  
            *dbright  = dv[j] + (Cq+shift-temp[j])*(dv[j+1]-dv[j])/(temp[j+1]-temp[j]); 
            *dbright *= -0.4 * log(10.0) * *bright; 
        } else if (phot[id] == 4) {   /* R photometric band */ 
            *bright   = rr[j] + (Cq+shift-temp[j])*(rr[j+1]-rr[j])/(temp[j+1]-temp[j]); 
            *bright   = pow(10.0, -0.4 * (*bright-mag0));  
            *dbright  = dr[j] + (Cq+shift-temp[j])*(dr[j+1]-dr[j])/(temp[j+1]-temp[j]); 
            *dbright *= -0.4 * log(10.0) * *bright; 
        } else if (phot[id] == 5) {   /* I photometric band */ 
            *bright   = ii[j] + (Cq+shift-temp[j])*(ii[j+1]-ii[j])/(temp[j+1]-temp[j]); 
            *bright   = pow(10.0, -0.4 * (*bright-mag0));  
            *dbright  = di[j] + (Cq+shift-temp[j])*(di[j+1]-di[j])/(temp[j+1]-temp[j]); 
            *dbright *= -0.4 * log(10.0) * *bright; 
        }
    }
}


static void sample(float *I, double *lambda, int n, float Cq, int id) 
{
/*
    if (Cq == Cq0)      linear(lambda,I,n,lam[id],spec[id],nspec[id]);   
    else if (Cq == Cq1) linear(lambda,I,n,lam_s[id],spec_s[id],nspec_s[id]);
*/
    if (Cq == Cq0)      spl(lambda,I,n,lam[id],spec[id],y2[id],nspec[id]);   
    else if (Cq == Cq1) spl(lambda,I,n,lam_s[id],spec_s[id],y2_s[id],nspec_s[id]);
    else { 
        printf("This should never happen\n"); 
        exit(1); 
    } 
}


void linear(double *x, float *y, int npt, float *lam, float *data, int nspec)
{
    int   k, xin, l, j;
    float slope;
 
    for (k = 0; k < npt; k++) {
        xin = 0; l = nspec-1;
        while (l-xin > 1) {
            j = (l + xin)/2;
            if (lam[j] > x[k]) l   = j;
            else               xin = j;
        }
        slope = (x[k]-lam[xin]) / (lam[xin+1]-lam[xin]);
        if (slope < 0.0)      slope = 0.0; 
        else if (slope > 1.0) slope = 1.0; 
        y[k] = (data[xin+1]-data[xin])*slope + data[xin];
    }
}


static void spl(double *x, float *y, int npt, float *lam, float *data, float *yy, int nspec)
{
    int   k, xin, l, j;
    float step, a, b;
 
    for (k = 0; k < npt; k++) {
        xin = 0; l = nspec-1;
        while (l-xin > 1) {
            j = (l + xin)/2;
            if (lam[j] > x[k]) l   = j;
            else               xin = j;
        }
        step = lam[xin+1]-lam[xin];
        a = (lam[xin+1]-x[k])/step;
        b = (x[k]-lam[xin])/step;
        if (b < 0.0) { 
            b = 0.0; a = 1.0; 
        } 
        if (a < 0.0) { 
            a = 0.0; b = 1.0; 
        } 
        y[k] = a*data[xin] + b*data[xin+1] +
             ((a*a*a-a)*yy[xin] + (b*b*b-b)*yy[xin+1])*step*step/6.0;
    }
}


int locate(float x, float array[], int n)
{
    int klo, khi, k, ascnd;
 
    klo = 0;
    khi = n-1;
    ascnd = (array[n-1] > array[0]);
    while (khi-klo > 1) {
        k = (khi + klo)/2;
        if (x > array[k] == ascnd)
            klo = k;
        else
            khi = k;
    }
    return klo;
}


static int read_table(char filename[], int batch)
{
    int   nl, j, k, n, *indx; 
    float tmp, *mult, *mult_s; 
    char  name[80]; 
    FILE  *fp; 

    printf("Name of file to read observed profile table from : ");
    scanf("%s", filename);
    if (batch == 1) printf("%s\n",filename);
    
    fp = fopen(filename,"r");
    if (fp == NULL) return 1;

    fscanf(fp, "%d", &nl); 

    if (nl) { 
        indx = ivector(0, nl-1); 
        mult = vector(0, nl-1); 
        mult_s = vector(0, nl-1); 
    } 

    for (j = 0; j < nl; j++) { 
        fscanf(fp, "%s", name); 
        indx[j] = -1; 
        for (k = 0; k < NUM_LINES; k++) { 
            if ((!strcmp(name,line_names[k])) && (!mod[k]) && (!phot[k])) { 
                mod[k] = 1; 
                indx[j] = k; 
                continue; 
            } 
        } 
    } 
    for (j = 0; j < NUM_LINES; j++)  
        if ((!mod[j]) && (!phot[j])) line_names[j] = empty; 

    lam    = (float **) svector(NUM_LINES,(int) sizeof(float *)); 
    lam_s  = (float **) svector(NUM_LINES,(int) sizeof(float *)); 
    spec   = (float **) svector(NUM_LINES,(int) sizeof(float *)); 
    spec_s = (float **) svector(NUM_LINES,(int) sizeof(float *)); 
    y2     = (float **) svector(NUM_LINES,(int) sizeof(float *));
    y2_s   = (float **) svector(NUM_LINES,(int) sizeof(float *));
    nspec   = ivector(0, NUM_LINES-1); 
    nspec_s = ivector(0, NUM_LINES-1); 
    for (j = 0; j < NUM_LINES; j++) nspec_s[j] = nspec[j] = 0; 
    for (j = 0; j < nl; j++) 
        if (indx[j] >= 0) { 
            fscanf(fp,"%d",nspec+indx[j]); 
            lam[indx[j]]  = vector(0, nspec[indx[j]]-1); 
            spec[indx[j]] = vector(0, nspec[indx[j]]-1); 
            y2[indx[j]]   = vector(0, nspec[indx[j]]-1); 
            for (k = 0; k < nspec[indx[j]]; k++) fscanf(fp,"%f",lam[indx[j]]+k); 
            for (k = 0; k < nspec[indx[j]]; k++) fscanf(fp,"%f",spec[indx[j]]+k); 
            spline(lam[indx[j]],spec[indx[j]],nspec[indx[j]],0.0,0.0,y2[indx[j]]);

            fscanf(fp,"%d",nspec_s+indx[j]); 
            lam_s[indx[j]]  = vector(0, nspec_s[indx[j]]-1); 
            spec_s[indx[j]] = vector(0, nspec_s[indx[j]]-1); 
            y2_s[indx[j]]   = vector(0, nspec_s[indx[j]]-1); 
            for (k = 0; k < nspec_s[indx[j]]; k++) fscanf(fp,"%f",lam_s[indx[j]]+k); 
            for (k = 0; k < nspec_s[indx[j]]; k++) fscanf(fp,"%f",spec_s[indx[j]]+k); 
            spline(lam_s[indx[j]],spec_s[indx[j]],nspec_s[indx[j]],0.0,0.0,y2_s[indx[j]]);
        } else { 
            nspec[indx[j]] = nspec_s[indx[j]] = 0; 
            fscanf(fp,"%d",&n); 
            for (k = 0; k < 2*n; k++) fscanf(fp,"%*f"); 
            fscanf(fp,"%d",&n); 
            for (k = 0; k < 2*n; k++) fscanf(fp,"%*f"); 
        } 
 

    for (j = 0; j < nl; j++) 
        if (indx[j] >= 0) {
            printf("Multiplication factors for line %s : ", line_names[indx[j]]); 
            scanf("%f %f", mult+j, mult_s+j); 
            if (batch) printf("%f %f\n", mult[j],mult_s[j]); 
        } 
    
    for (j = 0; j < nl; j++)  
        if (indx[j] >= 0) { 
            for (k = 0; k < nspec[indx[j]]; k++) 
                spec[indx[j]][k] = 1.0-mult[j]*(1.0-spec[indx[j]][k]);
            for (k = 0; k < nspec_s[indx[j]]; k++) 
                spec_s[indx[j]][k] = 1.0-mult_s[j]*(1.0-spec_s[indx[j]][k]);
        } 
 
    fclose(fp); 
    if (nl) free_vector(mult_s,  0, nl-1); 
    if (nl) free_vector(mult,  0, nl-1); 
    if (nl) free_ivector(indx, 0, nl-1); 

    return 0; 
}
