/*

This package contains routines that look after the gridding of the
star.  They are called by most of the other doppler imaging routines.

The information about the number and location of the latitude rings and their
division into longitude sectors is kept in static variables where it can
later be used by other routines. The static variables that will hold
the information between function calls are defined below.
*/

#define COMPILING_GRID
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "structures.h"
#include "constants.h"
#include "mem_alloc.h"
#include "grid.h"
#include "bessel.h"

#define MAX_RINGS 100
#define MRP1      (MAX_RINGS+1)

static int   rings, sectors[MAX_RINGS];
static float incl, veqt, th[MAX_RINGS], thb[MRP1], dph[MAX_RINGS], diffrot[2], fac;


/*
The routine init_grid() initialises the grid variables to give a total of
ngrid cells distributed as evenly as possible over the visible portion of
the star.  The variable inc is the inclination of the stars polar axis and
ve is the stars equatorial rotation velocity.
*/

void init_grid(ngrid,inc,ve)
int ngrid;
float inc,ve;
{
    int    lring,uring,temp,i,j,jmax,nmax,ni,ntot;
    double tempr1,tempr2;


    /* First we must find out how many grid cells to use */

    lring = uring = sqrt((double) ngrid);  /* Initialise lower and upper bounds */
    while ((temp = sum_cells(lring,inc)) > ngrid)
        lring /= 2; /* Find lower bound */
    while ((temp = sum_cells(uring,inc)) < ngrid)
        uring *= 2; /* Find upper bound */
    rings = opt_ring(ngrid,lring,uring,inc);  /* Pick the optimum number of rings */
    if (rings > MAX_RINGS) {
        fprintf(stderr,"Error in init_grid() : no of rings exceeds MAX_RINGS");
        exit(0);
    }

    /* Now we allocate longitude sectors to each ring */

    temp = sum_cells(rings,inc) - ngrid;
    ntot=0;
    for (j = 0; j < rings; j++) {
        ni = (float) sectors[j]*temp/(float)(ngrid+temp)+0.5;
        if (ni > temp-ntot) ni = temp-ntot;
        ntot += ni;
        sectors[j] -= ni;
    }
    temp -= ntot;
    for (i = 1; i <= temp; i++) {
        jmax = nmax = 0;
        for (j = 0; j < rings; j++) {
            if (sectors[j] > nmax) {
                jmax = j;
                nmax = sectors[j];
            }
        }
        sectors[jmax] -= 1;
    }

    /* Now we initialise the remaining static variables */

    incl = inc;
    veqt = ve;
    tempr1 = sin(inc) + 1.0;
    for (i = 1; i <= rings; i++) {
        tempr2 = tempr1*((float)i-0.5)/(float)rings-1.0;
        th[i-1] = PIon2 - inc + asin(tempr2);  /* Latitude, $\theta$, of each ring */
        tempr2 = tempr1*((float)i-1.0)/(float)rings-1.0;
        thb[i-1] = PIon2 - inc + asin(tempr2);  /* Latitude of the lower boundary of each ring */
        dph[i-1] = 2.0*PI/(float)sectors[i-1]; /* The width of the longitude
                                sectors, constant in each ring */

printf("%d %f %d\n", i-1, th[i-1], sectors[i-1]); 

    }
    thb[rings] = PIon2;  /* This is the upper boundary of the last ring which is always $\pi/2$ */
}


void init_diffrot(int batch)
{
    float beta, gamma; 

    /* sets coefficients for differential rotation, equal to that of ABDor in Dec 95 */ 
    /* Note: normally diffrot[0] and diffrot[1] should be of opposite sign, 
             diffrot[1] > 0.0 means equator rotating faster than pole */ 
/*
    diffrot[0] = -0.003109; diffrot[1] = 0.004614;   
*/

    /* values for rx j1508.5-44.23 in units of abdor*/ 
/*
    diffrot[0] = -0.000769; diffrot[1] = 0.002846;   
*/

    printf("Beta and gamma for differential rotation (0 for default) : "); 
    scanf("%f %f", &beta, &gamma);
/*
    if (beta < 0.0) beta = 0.0; if (gamma < 0.0) gamma = 0.0; 
*/
    if (batch) printf("%g %g\n\n",beta, gamma);

    diffrot[0] = beta; 
    diffrot[1] = gamma; 
    fac  = 0.0; /* take into account effect of diffrot on profile shape */ 

    veqt /= 1.0-fac*diffrot[0]; 
}


/*
The routine opt_ring() performs a root search (using the False Position
Method) to find the optimal number of latitude rings.  The returned number is
the smallest number of rings which which satisfies or exceeds the required
number of cells ngrid. The variables lring and uring must bracket the
root.
*/

#define  MAXIT  30  /* Controls the maximum number of iterations */

static int opt_ring(ngrid,lring,uring,inc)
int ngrid,lring,uring;
float inc;  /* Inclination angle of the stars rotation axis */
{
    int rl,ru,fl,fu,dr,i,rnew,fnew,converged;

    rl = lring;
    ru = uring;
    fl = sum_cells(rl,inc) - ngrid;
    fu = sum_cells(ru,inc) - ngrid;
    if (fl == 0) {
        return rl;
    }
    if(fu == 0) {
        return ru;
    }
    dr = ru-rl;
    converged = 0;
    for (i = 1; i <= MAXIT; i++) {
        rnew = (int)(rl+(float)dr*(float)fl/(float)(fl-fu)+0.999);
        if (rnew >= ru || rnew <= rl)
            rnew = (int)((rl+ru)/2.0);
        fnew = sum_cells(rnew,inc) - ngrid;
        if (fnew < 0) {
            rl = rnew;
            fl = fnew;
        } else {
            ru = rnew;
            fu = fnew;
        }
        dr = ru-rl;
        if (dr <= 1 || fnew == 0) {  /* Routine has converged */
            converged = 1;
            break;
        }
    }
    if (!converged)
        fprintf(stderr,"Error in opt_ring() : exceeded maximum iterations");
    if (fnew == 0)
        return rnew;
    else
        return ru;
}


/*
The routine sum_cells() calculates the total number of grid cells needed
if ring latitude rings are used.  The number of sectors per ring is
calculated to make the sectors at longitude zero appear approximately square
to the observer.
*/

static int sum_cells(ring,inc)
int ring;  /* Number of latitude rings in the trial configuration */
float inc; /* Inclination angle of the stars rotation axis */
{
    int    i,ni,sum;
    double theta1,theta2,dphi,tempr1,tempr2;

    rings = ring;
    if (rings <= 1) {
        sectors[0] = 1;
        return rings;
    }
    sum = 0;
    for (i = 1; i <= ring; i++) {
        tempr1 = (sin(inc) + 1.0)/rings;
        tempr2 = tempr1*((float)i-1.0) - 1.0;
        theta1 = PIon2 - inc + asin(tempr2);
        tempr2 = tempr1*((float)i) - 1.0;
        theta2 = PIon2 - inc + asin(tempr2);
        tempr2 = 2.0*(theta2-theta1)*sin(inc) - 
                      cos(2.0*theta2+inc)+cos(2.0*theta1+inc);
        dphi = 4.0*tempr1*tempr1/tempr2;
        ni = (int) (2.0*PI/dphi + 0.5);
        if (i <= MAX_RINGS)
            sectors[i-1] = ni;
        sum += ni;
    }
    return sum;
}


/*
One further initialisation procedure is needed to set the angle between the
polariod axis and the projected rotation axis. This is done in a seperate
routine to the main initialisation above so that it can be easily reset
many times per calculation if necessary. The orientation of the polaroid
axis only affects linear polarisation measurements.
*/

static float master_polariser = 0.0;

void set_polaroid(angle)
float angle;
{
    master_polariser = angle;
}


/*
The routine geometry() communicates the grid information to the other imaging
and synthesis routines.  It returns all the relevant information about each
grid point at the given phase.  The variable n contains the number of grid
points.  mu is set to the cosine of the limb angle at that grid point.
parea is set to the projected area of the grid cell.  If the cell isn't in the
visible hemisphere, parea is set to zero.  vel is the rotational
velocity relative to the observer at the grid point.  vel is -ve if the
point is approaching and +ve if receeding.
*/

void geometry(phase,n,geom,magnetic,Br,Bt,Bp,B,gamma,chi,derivs)
float phase;  /* Rotational phase $\Phi$ */
int n;        /* Number of grid cells */
cell_info_t geom[];   /* Array to hold the returned geometrical data */
int magnetic;         /* Switch to indicate whether to do magnetic field calculations */
float Br[],Bt[],Bp[]; /* Magnetic field components */
float B[],gamma[],chi[]; /* Magnetic field transformed to the Stokes frame */
cell_deriv_t derivs[];   /* Array to hold the field angle derivatives */
{
    int i,j,k;
    double theta,pphi,ct,st,cp,sp,ci,si,btmp,tmp,c_r,c_t,c_p,c_rr,c_tt,c_pp,B2;
    double phi1,phi2,cp1,cp2,sp1,sp2,mu1,mu2,th1,th2,st1,st2,ct1,ct2,rB3;

    ci = cos(incl);
    si = sin(incl);
    k = 0;   /* Index of the current grid cell */
    for (i = 0; i < rings; i++) {  /* Loop over rings */
        th2   = thb[i+1];  ct2 = cos(th2);  st2 = sin(th2);  /* $\theta$ at northern edge */
        for (j = 0; j < sectors[i]; j++) { /* At each ring move along the sectors */
              /* Calculate the cell location */
            theta = th[i]; ct  = cos(theta); st  = sin(theta);  /* theta */
            th1  = thb[i]; ct1 = cos(th1);   st1 = sin(th1);    /* theta at southern edge */
            pphi = (float) j*dph[i] + 2.0*PI*phase;    /* \phi */
/* correction to phi to account for differential rotation, jfd, 15jun98 */ 
pphi -= 2.0*PI*phase * (diffrot[0] + diffrot[1]*st*st);  
            for (; (pphi <= -PI) || (pphi > PI+EPS); ) {  /* Convert phi to range -pi < phi <= pi */
                if (pphi <= -PI) pphi += 2.0*PI; 
                if (pphi > PI)   pphi -= 2.0*PI; 
            }
            cp = cos(pphi);  sp = sin(pphi);
            phi1 = pphi-dph[i]/2.0; phi2 = pphi+dph[i]/2.0;  /* Longitudes of the cell edges */

              /* Calculate the projected area */
            cp1 = cos(phi1);  sp1 = sin(phi1);  mu1 = ct2*si*cp1+st2*ci;  /* Northeast corner */
            cp2 = cos(phi2);  sp2 = sin(phi2);  mu2 = ct2*si*cp2+st2*ci;  /* Northwest corner */
            if ((mu1 > 0.0) || (mu2 > 0.0)) {  /* Is the cell visible? */
                if (mu1 < 0.0) {
                    phi1 = -PIon2-asin(st2/ct2*ci/si); sp1 = sin(phi1); 
                    pphi = 0.5*(phi1+phi2); cp = cos(pphi); sp = sin(pphi); 
                } else if (mu2 < 0.0) {
                    phi2 = PIon2+asin(st2/ct2*ci/si);  sp2 = sin(phi2); 
                    pphi = 0.5*(phi1+phi2); cp = cos(pphi); sp = sin(pphi); 
                }
                if (ct1*si*cp+st1*ci < 0.0) { /* Is the bottom of the cell invisible? */
                    th1 = - atan(cp*si/ci); ct1 = cos(th1); st1 = sin(th1); 
                    tmp = -(cos(th1+incl)+cos(th2+incl))/2.0; 
                    theta = PIon2-incl+asin(tmp); ct = cos(theta); st = sin(theta); 
                }
                geom[k].parea = (si*(sp2-sp1)*((st2*ct2-st1*ct1)+th2-th1) - 
                    ci*(phi2-phi1)*(ct2*ct2-ct1*ct1))/(2.0*PI); /* The formula for $A_p$ */
                if (geom[k].parea <= 0.0) geom[k].parea = 0.0;  /* Oops! our approximation has gone astray */
            } else 
                geom[k].parea = 0.0;  /* Cell is invisible */

            if (geom[k].parea > 0.0) {
                geom[k].mu = ct*si*cp+st*ci;  /* mu */
                geom[k].vel = veqt*sp*si*ct;  /* v_parallel */
/* correction to phi to account for differential rotation, jfd, 05aug02 */ 
geom[k].vel = veqt*sp*si*ct*(1.0+fac*(diffrot[1]*(ct*ct-1.0)-diffrot[0]));  /* v_parallel */

                if (magnetic) {  /* We need to calculate the magnetic field stuff */
                    btmp = sqrt(Br[k]*Br[k] + Bt[k]*Bt[k] + Bp[k]*Bp[k]); /* $B$ */
                    B[k] = btmp;
                    B2 = btmp*btmp;  /* Square of the magnitude of the magnetic field */
                    if (btmp <= 0.0) {
			btmp = rB3 = 1.0;
                    } else
		        rB3 = 1.0/(B2*btmp);

                    c_r = st*ci+ct*cp*si;  c_t = -sp*si;  c_p = ct*ci-st*cp*si; /* ${\bf e}_3$ components */
                    tmp = Br[k]*c_r + Bt[k]*c_t + Bp[k]*c_p;
                    gamma[k] = acos(tmp/btmp); /* $\gamma$ */
                    derivs[k].dcgdBr = (c_r*B2 - tmp*Br[k])*rB3; /* $\partial(\cos\gamma) / \partial B_r$ */
                    derivs[k].dcgdBt = (c_t*B2 - tmp*Bt[k])*rB3; /* $\partial(\cos\gamma) / \partial B_t$ */
                    derivs[k].dcgdBp = (c_p*B2 - tmp*Bp[k])*rB3; /* $\partial(\cos\gamma) / \partial B_p$ */
                    c_r  = -ct*sp;          c_t  = -cp;    c_p  = st*sp; /* ${\bf e}_2$ components */
                    c_rr = st*si-ct*cp*ci;  c_tt = sp*ci;  c_pp = ct*si+st*cp*ci; /* ${\bf e}_1$ components */
                    tmp  = Br[k]*c_r + Bt[k]*c_t + Bp[k]*c_p;     /* ${\bf e}_2$ field */
                    btmp = Br[k]*c_rr + Bt[k]*c_tt + Bp[k]*c_pp;  /* ${\bf e}_1$ field */
		    if (tmp == 0.0 && btmp == 0.0)
                        chi[k] = master_polariser; /* $\chi$ */
		    else
                        chi[k] = atan2(tmp,btmp) + master_polariser; /* $\chi$ */
                    B2 = tmp*tmp + btmp*btmp;  /* Square of the magnitude of the ``horizontal'' field */
                    if (B2 == 0) B2 = 1;   /* Prevent divide by zero errors.
                                       Makes the derivatives zero when $B_{\rm horiz}$ vanishes */
                    derivs[k].dchdBr = (c_r*btmp - c_rr*tmp) / B2;   /* $\partial\chi / \partial B_r$ */
                    derivs[k].dchdBt = (c_t*btmp - c_tt*tmp) / B2;   /* $\partial\chi / \partial B_t$ */
                    derivs[k].dchdBp = (c_p*btmp - c_pp*tmp) / B2;   /* $\partial\chi / \partial B_p$ */
                }
            } 
            k++;
        }
    }
}


/*
Routine returns the area of each grid cell.
*/

void areas(ngrid,area)
int ngrid;
float area[];
{
    int   i,j,k;
    float aa;

    k = 0;
    for (i = 0; i < rings; i++) {
	aa = (sin(thb[i+1]) - sin(thb[i]))*dph[i]*ngrid/(4.0*PI);
        for (j = 0; j < sectors[i]; j++) {
            area[k] = aa;
            k++;
        }
    }
}

/*
Routine returns the location of each grid point at phase phase.
ngrid is the number of grid points.  theta is the latitude of
the grid point and phi is its longitude.
*/

void locations(phase,ngrid,theta,phi)
float phase;
int ngrid;
float theta[],phi[];
{
    int   i,j,k;
    float pphi;

    k = 0;
    for (i = 0; i < rings; i++) {
        for (j = 0; j < sectors[i]; j++) {
            theta[k] = th[i];
            pphi     = (float)j*dph[i] + 2.0*PI*phase;
            for (; (pphi <= -PI) || (pphi > PI+EPS); ) {  /* Convert phi to range
                                                           -pi < phi <= pi */
                if (pphi <= -PI) pphi += 2.0*PI; 
                if (pphi > PI)   pphi -= 2.0*PI; 
            }
	    phi[k] = pphi;
            k++;
        }
    }
}


/*
Routine returns the grid index number n of the grid cell which
contains the point with stellar coordinates $(\theta,\phi)$ at phase
phase.
*/

void find_grid_loc(theta,phi,n,phase)
float theta,phi,phase;
int *n;
{
    int i,j;

    if (theta < -incl) {
        *n = 0;
        return;
    }

    *n = i = 1;

    while (thb[i-1] <= theta && i < rings) {
        *n += sectors[i-1];
        i++;
    }

    if (i > 1) {
        if (thb[i-1] > theta) {
            i -= 1;
            *n -= sectors[i-1];
        }
    } else
        return;

    phi = phi-2.0*PI*phase;

    while (phi < -dph[i-1]/2.0)
        phi += (2.0*PI);
    while (phi > 2.0*PI - dph[i-1]/2.0)
        phi -= (2.0*PI);

    j = 1;
    while (phi > ((j++)-0.5)*dph[i-1])
        *n += 1;
    return;
}


void legendre(float ***L, int ngrid, int lmax, int potential, float alpha)
{
    int   i, j, k, l, m, n, ntot, kk; 
    float ct,st,cp,sp,phi,theta,pmm,pmmp1,plm,plmp1,plmm1,plmm2,*ylm,*J,*Y,
          A,B,r,s,limit,*Y1,*Y2,tmp;
printf("here\n"); 

    limit = 40.0; kk = 1000; 
    if ((potential == 3) || (potential == 5)) { 
        Y = vector(0, lmax); 
        J = vector(0, lmax); 
        bessel(J,Y,lmax,alpha,limit); 
    } else if (potential == -2) { 
        Y1 = vector(0, lmax); 
        Y2 = vector(0, lmax); 
        J = vector(0, lmax); 
        bessel(J,Y1,lmax,alpha/kk,limit); 
        bessel(J,Y2,lmax,alpha*kk,limit); 
        free_vector(J,0,lmax); 
    } 
    n = 0; ntot = (lmax+1)*(lmax+2)/2-1; 
    ylm = vector(0, ntot-1); 
    for (i = 0; i < rings; i++) {  
        theta = th[i]; st  = cos(theta); ct  = sin(theta);  
        pmm = 1.0;  
        for (m = 0; m <= lmax; m++) { 
            if (m > 0) ylm[(m+1)*(m+2)/2-2] = pmm; 
            if (m < lmax) {
                pmmp1 = ct*(2*m+1)*pmm; 
                ylm[(m+1)*(m+2)/2+m-1] = pmmp1; 
            } 
            plmm2 = pmm; plmm1 = pmmp1; 
            if (m < lmax-1)  
                for (l = m+2; l <= lmax; l++) { 
                    plm=(ct*(2*l-1)*plmm1-(l+m-1)*plmm2)/(l-m); 
                    ylm[l*(l+1)/2+m-1]=plm; 
                    plmm2=plmm1;  
                    plmm1=plm; 
                } 
            pmm *= -st*(2*m+1);
        } 
        for (j = 0; j < sectors[i]; j++) { 
            phi = (float) j*dph[i];    
            for (m = 0; m <= lmax; m++) { 
                sp = sin(phi*m); cp = cos(phi*m); 
                for (l = m; l <= lmax; l++) {
                    if (l == 0) continue; 
                    plm = ylm[l*(l+1)/2+m-1]; 
                    if (l > m) plmp1 = ylm[l*(l+1)/2+m]; 
                    else       plmp1 = 0.0; 
                    A = plm*m/st/(l+1); 
                    B = (ct*m*plm/st+plmp1)/(l+1); 
                    if ((potential == 3) || (potential == 5)) { 
                        r = Y[l-1]*alpha; s = Y[l]; if (alpha <= limit) r *= 0.5*alpha/l; 
                        L[2][l*(l+1)+2*m-2][n] =  (s-r/l)*B*cp + s*alpha/l*A*sp; 
                        L[2][l*(l+1)+2*m-1][n] =  (s-r/l)*B*sp - s*alpha/l*A*cp;
                        L[1][l*(l+1)+2*m-2][n] =  (s-r/l)*A*sp - s*alpha/l*B*cp;
                        L[1][l*(l+1)+2*m-1][n] = -(s-r/l)*A*cp - s*alpha/l*B*sp;
                        L[0][l*(l+1)+2*m-2][n] =  s*plm*cp; 
                        L[0][l*(l+1)+2*m-1][n] =  s*plm*sp; 
                        if (potential == 3) { 
                            r = J[l-1]*alpha; s = J[l]; if (alpha <= limit) r /= 0.5*alpha/l; 
                            L[5][l*(l+1)+2*m-2][n] =  (s-r/l)*B*cp + s*alpha/l*A*sp; 
                            L[5][l*(l+1)+2*m-1][n] =  (s-r/l)*B*sp - s*alpha/l*A*cp;
                            L[4][l*(l+1)+2*m-2][n] =  (s-r/l)*A*sp - s*alpha/l*B*cp;
                            L[4][l*(l+1)+2*m-1][n] = -(s-r/l)*A*cp - s*alpha/l*B*sp;
                            L[3][l*(l+1)+2*m-2][n] =  s*plm*cp; 
                            L[3][l*(l+1)+2*m-1][n] =  s*plm*sp; 
                        } 
                    } else if (potential == -1) { 
/*
                        L[5][l*(l+1)+2*m-2][n] =  B*cp*(l+1)/l; 
                        L[5][l*(l+1)+2*m-1][n] =  B*sp*(l+1)/l;
                        L[4][l*(l+1)+2*m-2][n] =  A*sp*(l+1)/l; 
                        L[4][l*(l+1)+2*m-1][n] = -A*cp*(l+1)/l; 
                        L[3][l*(l+1)+2*m-2][n] = -plm*cp; 
                        L[3][l*(l+1)+2*m-1][n] = -plm*sp; 
                        L[2][l*(l+1)+2*m-2][n] =  B*cp; 
                        L[2][l*(l+1)+2*m-1][n] =  B*sp;
                        L[1][l*(l+1)+2*m-2][n] =  A*sp; 
                        L[1][l*(l+1)+2*m-1][n] = -A*cp; 
                        L[0][l*(l+1)+2*m-2][n] = plm*cp; 
                        L[0][l*(l+1)+2*m-1][n] = plm*sp; 

                        L[5][l*(l+1)+2*m-2][n] =  B*cp; 
                        L[5][l*(l+1)+2*m-1][n] =  B*sp;
                        L[4][l*(l+1)+2*m-2][n] =  A*sp; 
                        L[4][l*(l+1)+2*m-1][n] = -A*cp; 
                        L[3][l*(l+1)+2*m-2][n] = -plm*cp; 
                        L[3][l*(l+1)+2*m-1][n] = -plm*sp; 
                        L[2][l*(l+1)+2*m-2][n] =  B*cp*(l+1)/l; 
                        L[2][l*(l+1)+2*m-1][n] =  B*sp*(l+1)/l;
                        L[1][l*(l+1)+2*m-2][n] =  A*sp*(l+1)/l;
                        L[1][l*(l+1)+2*m-1][n] = -A*cp*(l+1)/l;
                        L[0][l*(l+1)+2*m-2][n] = plm*cp; 
                        L[0][l*(l+1)+2*m-1][n] = plm*sp; 
*/
                        L[5][l*(l+1)+2*m-2][n] =  B*cp; 
                        L[5][l*(l+1)+2*m-1][n] =  B*sp;
                        L[4][l*(l+1)+2*m-2][n] =  A*sp; 
                        L[4][l*(l+1)+2*m-1][n] = -A*cp; 
                        L[3][l*(l+1)+2*m-2][n] = -plm*cp*l/(l+1); 
                        L[3][l*(l+1)+2*m-1][n] = -plm*sp*l/(l+1); 
                        L[2][l*(l+1)+2*m-2][n] =  B*cp; 
                        L[2][l*(l+1)+2*m-1][n] =  B*sp;
                        L[1][l*(l+1)+2*m-2][n] =  A*sp;
                        L[1][l*(l+1)+2*m-1][n] = -A*cp;
                        L[0][l*(l+1)+2*m-2][n] = plm*cp; 
                        L[0][l*(l+1)+2*m-1][n] = plm*sp; 
                    } else if (potential == -2) { 
                        tmp = alpha/kk; 
                        r = Y1[l-1]*tmp; s = Y1[l]; if (tmp <= limit) r *= 0.5*tmp/l;
                        L[2][l*(l+1)+2*m-2][n] =  (s-r/l)*B*cp + s*tmp/l*A*sp; 
                        L[2][l*(l+1)+2*m-1][n] =  (s-r/l)*B*sp - s*tmp/l*A*cp;
                        L[1][l*(l+1)+2*m-2][n] =  (s-r/l)*A*sp - s*tmp/l*B*cp;
                        L[1][l*(l+1)+2*m-1][n] = -(s-r/l)*A*cp - s*tmp/l*B*sp;
                        L[0][l*(l+1)+2*m-2][n] =  s*plm*cp; 
                        L[0][l*(l+1)+2*m-1][n] =  s*plm*sp; 
                        tmp = alpha*kk; 
                        r = Y2[l-1]*tmp; s = Y2[l]; if (tmp <= limit) r *= 0.5*tmp/l;
                        L[5][l*(l+1)+2*m-2][n] =  (s-r/l)*B*cp + s*tmp/l*A*sp; 
                        L[5][l*(l+1)+2*m-1][n] =  (s-r/l)*B*sp - s*tmp/l*A*cp;
                        L[4][l*(l+1)+2*m-2][n] =  (s-r/l)*A*sp - s*tmp/l*B*cp;
                        L[4][l*(l+1)+2*m-1][n] = -(s-r/l)*A*cp - s*tmp/l*B*sp;
                        L[3][l*(l+1)+2*m-2][n] =  s*plm*cp; 
                        L[3][l*(l+1)+2*m-1][n] =  s*plm*sp; 
                    } else { 
                        L[2][l*(l+1)+2*m-2][n] =  B*cp; 
                        L[2][l*(l+1)+2*m-1][n] =  B*sp;
                        L[1][l*(l+1)+2*m-2][n] =  A*sp; 
                        L[1][l*(l+1)+2*m-1][n] = -A*cp; 
                        L[0][l*(l+1)+2*m-2][n] = plm*cp; 
                        L[0][l*(l+1)+2*m-1][n] = plm*sp; 
                    } 
/*
printf("%d\n", l*(l+1)+2*m-1); 
*/
                }
            } 
            n++; 
        } 
    } 
    free_vector(ylm,0,ntot-1); 
    if ((potential == 3) || (potential == 5)) { 
        free_vector(J,0,lmax); 
        free_vector(Y,0,lmax); 
    } else if (potential == -2) { 
        free_vector(Y1,0,lmax); 
        free_vector(Y2,0,lmax); 
    } 
}

