/* This routine uses tables with synthetic profiles obtained from Kurucz 
SYNTHE code.  */ 

#include <stdio.h>
#include <math.h>
#include "structures.h"
#include "mem_alloc.h"
#include "constants.h"
#include "ubv.h"

static float Cq0;  /* default value for Cq */ 
static float Cq1;  /* default value for Cq in spots (when reconstructing ff's) */ 
static int   ff;   /* indicating filling factor reconstruction */ 
static int   ap;   /* indicating temperature or abundance reconstruction */ 
static int nbr, nmu, nlam; 
static float *mmu, *br, ****prof, ****dprof; 
static double *laml, *lamu; 

/*  We use static variables to hold the line information.  */

#define NUM_LINES 14
static char *line_names[NUM_LINES] = { 
                                       "Fe1_393",
                                       "Ca_K",
                                       "Al_394",
                                       "Cr2_404",
                                       "Fe1_492",
                                       "Cr1_492",
                                       "Fe2_492",
                                       "Ti_518",
                                       "Fe_550",
                                       "Fe_643",
		                       "Ca_644",
                                       "O_triplet",
		                       "B_band",
		                       "V_band" };

static int line_ids[NUM_LINES] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13};

static int phot[NUM_LINES] = {0,0,0,0,0,0,0,0,0,0,0,0,2,3};  
    /* 0 is for a spectral line, 1,2,3 are for UBV bands */ 

static int mod[NUM_LINES] = {0,0,0,0,0,0,0,0,0,0,0,0};  
    /* mod is modified when reading kurucz tables */ 

static int lin[NUM_LINES] = {0,0,0,0,0,0,0,0,0,0,0,0};

static int mag_opt[NUM_LINES] = {1,1,1,1,1,1,1,1,1,1,1,1}; 

static double l0[NUM_LINES]  = {          /* $\lambda_0$ */
                                 392.7920,
                                 393.3663,
                                 394.4006,
                                 403.7972,
                                 491.8993,
                                 492.2280,
                                 492.3927,
                                 518.8680,
                                 549.7516,
                                 643.0844,
			 	 643.9075,
                                 777.3500,
			 	 450.0000,
			 	 550.0000 }; 
static double eps[NUM_LINES] = {         /* $\epsilon$ */
                                 0.60, 
                                 0.60, 
                                 0.60, 
                                 0.45, 
                                 0.45, 
                                 0.45, 
                                 0.45, 
                                 0.65, 
                                 0.65,
				 0.65, 
				 0.30};
static double g[NUM_LINES] =   {         /* Lande g factor */
                                 0.00,
                                 0.00,
                                 0.00,
				 0.00, 
				 0.00, 
				 0.00, 
				 0.00, 
				 0.00, 
				 0.00, 
                                 0.00,
                                 0.00}; 
static double zeta_rt[NUM_LINES] = {     /* $\zeta$ */
                                 0.00,
                                 0.00,
				 0.00, 
                                 0.00,
				 0.00,
				 0.00,
				 0.00, 
				 0.00, 
				 0.00, 
				 0.00,
                                 0.00}; 


void init_spectral_lines(batch)
int batch;
{
    int   err;  
    char filename[80]; 
    int read_table(char filename[], int batch); 

    do {
        err = read_table(filename, batch); 
        if (err)
            if (batch == 1) {
                if (err == 1) printf("File not found.\n");
                else          printf("Wrong file.\n");
                exit(1);
            } else { 
                if (err == 1) printf("File not found.  Try again\n");
                else          printf("Wrong file.  Try again\n");
            } 
    } while (err);
}


void end_spectral_lines()
{
    if (nmu*nbr*nlam) free_matrix4D(prof,  0, NUM_LINES-1, 0, nmu-1, 0, nbr-1, 0, nlam-1);
    if (nmu*nbr*nlam) free_matrix4D(dprof, 0, NUM_LINES-1, 0, nmu-1, 0, nbr-1, 0, nlam-1);

    free_dvector(laml, 0, NUM_LINES-1); 
    free_dvector(lamu, 0, NUM_LINES-1); 
    if (nmu) free_vector(mmu,   0, nmu-1); 
    if (nbr) free_vector(br,    0, nbr-1); 
}


void get_spectral_lines(nlines,names,ids)
int *nlines,**ids;
char **names[];
{
    *nlines = NUM_LINES;
    *names = line_names;
    *ids   = line_ids;
}


/*
The following routine lets the main program find out some details about
each of the lines.
*/

void line_info(id,ll0,dll,broadened,linear)
int id;      /* Line ID number */
double *ll0;  /* Line centre wavelength */
double *dll;  /* Doppler width */
int *broadened, *linear;  /* Does the line model include the instrumental broadening */
{
    *ll0 = l0[id];
    *dll = 0.0;     /* Doppler width dependant on Cq */ 
    *broadened = 0;
    *linear = lin[id]; 
}


/*
The function line_centre() needs to return the centre wavelength of the
line being studied.
*/

double find_line_centre(id)
int id;
{
    return l0[id];
}


/*
The function find_macro_turb() needs to return the radial-tangential
macroturbulence parameter.  If you don't need to use macroturbulence
just return zero.  If you do need it you should look at the routines
doppler_spec() and doppler_deriv() to check how it is being
implimented.
*/

double find_macro_turb(id)
int id;
{
    return zeta_rt[id];
}


/*
This routine is used when a spectrum has to be calculated from an image.
You only need to calculate the Stokes parameters that you will use.  For
example, in straight Doppler imaging you only have to calculate Stokes I
using the continuum brightness C_q.
*/

void line_profile(id,I,Q,U,V,n,lambda,Cq,Cm,mu,B,gam,chi,magnetic)
int id;                /* Line ID number */
float I[],Q[],U[],V[];  /* Returned Stokes parameters */
int n;                /* Number of wavelengths to calculate */
double lambda[];        /* Wavelength $\lambda$ */
float Cq,Cm;          /* Continuum brightness */
float mu;              /* Cosine of the limb angle */
float B,gam,chi;     /* Magnetic field strength and orientation $(B,\gamma,\chi)$ */
int magnetic;          /* Switch is non-zero for a magnetic star */
{
    int    j;
    float *dI, *tmp; 
    float bright,dbright,bright_s,dbright_s,contin;
    void convert(float Cq, float *bright, float *dbright, int id); 
    void sample(float *I, float *dI, int deriv, double *lambda, int n, float Cq, 
                float mu, int id); 

    if (!ap) { 
        if (ff == 0)    
            convert(Cq, &bright, &dbright, id); 
        else { 
            convert(Cq0, &bright, &dbright, id); 
            convert(Cq1, &bright_s, &dbright_s, id); 
        }
    } else 
        bright = bright_s = 1.0; 

    if (!phot[id]) { /* Spectral line */
        if (magnetic) {
            printf("This is not the correct line profile routine to use\n"); 
            exit(1); 
        } else { 
            contin = (1.0-eps[id]+eps[id]*mu)/(1.0-eps[id]/3.0);
            if (ff == 0) {
                sample(I, dI, 0, lambda, n, Cq, mu, id); 
                for (j = 0; j < n; j++) 
                    I[j] *= contin*bright; 
            } else {
                tmp = vector(0, n-1); 
                sample(I, dI, 0, lambda, n, Cq0, mu, id); 
                sample(tmp, dI, 0, lambda, n, Cq1, mu, id); 
                for (j = 0; j < n; j++) 
                    I[j] = (Cq*bright*I[j]+(1.0-Cq)*bright_s*tmp[j])*contin; 
                free_vector(tmp, 0, n-1); 
            } 
        }

    } else { /* Photometry */
        contin = (1.0-eps[id]+eps[id]*mu)/(1.0-eps[id]/3.0);
        if (ff == 0) 
            I[0] = bright*contin; 
        else 
            I[0] = (Cq*bright+(1.0-Cq)*bright_s)*contin; 
    }
}


/*
This routine is an extended version of line_profile() used during the
image inversion process.  As well as returning the spectral profile you
also need to return the derivatives of the field profile with respect
to the image quantities you are fitting.
*/

void line_deriv(id,dII,dQQ,dUU,dVV,I,Q,U,V,n,lambda,Cq,Cm,mu,B,gam,chi,magnetic)
int id;
float **dII,**dQQ,**dUU,**dVV;  /* Arrays of profile derivatives */
float I[],Q[],U[],V[];  /* Returned Stokes parameters */
int   n;          /* Number of wavelength points */
double lambda[];    /* Wavelength $\lambda$ */
float Cq,Cm;        /* Continuum brightness */
float mu;           /* Cosine of the limb angle */
float B,gam,chi;    /* Magnetic field strength and orientation $(B,\gamma,\chi)$ */
int magnetic;       /* Switch is non-zero for a magnetic star */
{
    int    j;
    float *dI,*tmp; 
    float bright,dbright,bright_s,dbright_s,contin;
    void convert(float Cq, float *bright, float *dbright, int id); 
    void sample(float *I, float *dI, int deriv, double *lambda, int n, float Cq, 
                float mu, int id); 

    if (!ap) { 
        if (ff == 0)    
            convert(Cq, &bright, &dbright, id); 
        else { 
            convert(Cq0, &bright, &dbright, id); 
            convert(Cq1, &bright_s, &dbright_s, id); 
        }
    } else { 
        bright  = bright_s  = 1.0; 
        dbright = dbright_s = 0.0; 
    } 

    if (!phot[id]) { /* Spectral line */
        if (magnetic) {
            printf("This is not the correct line profile routine to use\n"); 
            exit(1); 
        } else { 
            contin = (1.0-eps[id]+eps[id]*mu)/(1.0-eps[id]/3.0);
            if (ff == 0) {
                dI = vector(0, n-1); 
                sample(I, dI, 1, lambda, n, Cq, mu, id); 
                for (j = 0; j < n; j++) {
                    I[j] *= contin*bright; 
                    dII[0][j] = dbright*I[j]/bright + contin*bright*dI[j];  
                }
                free_vector(dI,0,n-1);   
            } else {
                tmp = vector(0, n-1); 
                sample(I, dI, 0, lambda, n, Cq0, mu, id); 
                sample(tmp, dI, 0, lambda, n, Cq1, mu, id); 
                for (j = 0; j < n; j++) {
                    dII[0][j] = (bright*I[j]-bright_s*tmp[j])*contin; 
                    I[j] = (Cq*bright*I[j]+(1.0-Cq)*bright_s*tmp[j])*contin; 
                }
                free_vector(tmp, 0, n-1); 
            } 
        }

    } else { /* Photometry */
        contin = (1.0-eps[id]+eps[id]*mu)/(1.0-eps[id]/3.0);
        if (ff == 0) {
            I[0] = bright*contin; 
            dII[0][0] = dbright*contin; 
        } else { 
            I[0] = (Cq*bright+(1.0-Cq)*bright_s)*contin; 
            dII[0][0] = (bright-bright_s)*contin; 
        }
    }
}


void continuum(id,mu,Cq,Cm,cont,dCq,dCm,magnetic)
int id,magnetic;
float mu,Cq,Cm,*cont,*dCq,*dCm;
{
    float bright, dbright, bright_s, dbright_s, limb; 
    void convert(float Cq, float *bright, float *dbright, int id); 

    limb  = (1.0-eps[id]+eps[id]*mu)/(1.0-eps[id]/3.0);

    if (ap) { 
        *cont = limb;
        *dCq  = 0.0;
    } else { 
        if (ff == 0) {
            convert(Cq, &bright, &dbright, id); 
            *cont = bright*limb;
            *dCq  = dbright*limb; 
        } else { 
            convert(Cq0, &bright, &dbright, id); 
            convert(Cq1, &bright_s, &dbright_s, id); 
            *cont = (Cq*bright+(1.0-Cq)*bright_s)*limb; 
            *dCq  = (bright-bright_s)*limb; 
        }
    }
}


void image_info(Cm_used,qnames,ff_entropy,ff_rec,batch)
int *Cm_used,*ff_entropy,*ff_rec,batch;
char *qnames[];
{
    static char *temp = "temperature";
    static char *abun = "abundance";
    static char *fill = "filling factor";
    int rec; 

    *Cm_used = 0;

    printf("Type of image (0: unconstrained temperature; \n"); 
    printf("               1: dark-spot temperature; \n"); 
    printf("               2: temperature filling factor; \n"); 
    printf("               3: unconstrained abundance; \n");  
    printf("               4: abundance filling factor) : ");  
    scanf("%d", &rec); 
    printf("%d\n", rec); 

    *ff_entropy = *ff_rec = 0; 
    if (rec == 1) *ff_entropy = 1; 
    else if (rec == 3) *ff_entropy = -1; 
    else if ((rec == 2) || (rec == 4)) *ff_entropy = *ff_rec = 1; 
    ff = *ff_rec;  
    if (rec > 2) ap = 1; 
    else         ap = 0; 

    if (ff) 
        qnames[0] = fill;
    else { 
        if (ap) qnames[0] = abun;
        else    qnames[0] = temp;
    }
}


void image_defaults(dCq, dB, magnetic, Cm_used, qnames, batch) 
float *dCq, *dB; 
int magnetic, Cm_used, batch; 
char *qnames[]; 
{
    static char *temp = "temperature";
    static char *abun = "abundance";

    printf("Input default %s : ",qnames[0]);
    scanf("%f", dCq);
    if (batch) printf("%f\n",*dCq);
    if (ff) {
        if (ap) printf("Input default %s in photosphere : ", abun);
        else    printf("Input default %s in photosphere : ", temp);
        scanf("%f", &Cq0);
        if (batch) printf("%f\n",Cq0);
        if (ap) printf("Input default %s in spots : ", abun);
        else    printf("Input default %s in spots : ", temp);
        scanf("%f", &Cq1);
        if (batch) printf("%f\n",Cq1);
    } else
        Cq0 = *dCq;

    if (magnetic && Cm_used) {
        printf("What is the default %s : ",qnames[1]);
        scanf("%f",dB);
        if (batch) printf("%f\n",*dB);
    }
} 


void transform(Cq,Cm,ngrid,magnetic)
float Cq[],Cm[];
int ngrid,magnetic;
{
}


void untransform(Cq,Cm,ngrid,magnetic)
float Cq[],Cm[];
int ngrid,magnetic;
{
}


void transform_photometry(spec_info, nobs, I, mag0, wphot)
spec_info_t *spec_info; 
int nobs; 
float *I, mag0, wphot; 
{
    int j, n; 

    for (j = n = 0; j < nobs; n += spec_info[j].nspec, j++) 
        if (spec_info[j].photometric && spec_info[j].selected) {
            I[n] = pow(10.0, -0.4*(I[n]-mag0)); 
            spec_info[j].SN = I[n]*(pow(10.0, 0.4*spec_info[j].SN)-1.0)/sqrt(wphot); 
        }
}


void untransform_photometry(spec_info, nobs, I, mag0, wphot)
spec_info_t *spec_info; 
int nobs; 
float *I, mag0, wphot; 
{
    int j, n; 

    for (j = n = 0; j < nobs; n += spec_info[j].nspec, j++) 
        if (spec_info[j].photometric && spec_info[j].selected) {
            spec_info[j].SN = 2.5*log10(1.0+spec_info[j].SN*sqrt(wphot)/I[n]); 
            I[n] = mag0-2.5*log10(I[n]); 
        }
}


void convert(float Cq, float *bright, float *dbright, int id)
{
    float k, u, mag0, shift, fac; 
    int    j, j0; 
    int locate(float x, float array[], int n); 

    if (!phot[id]) { /* Spectral lines */ 
        k = exp(1.439e7/(l0[id]*Cq0))-1.0; 
        u = exp(1.439e7/(l0[id]*Cq))-1.0; 
        *bright  = k/u; 
        *dbright = k*1.439e7*(u+1.0)/(l0[id]*u*u*Cq*Cq); 
   } else {      /* Photometry */ 
        shift = 0.0; /* compensate for errors in the calibration table */ 
        j0 = locate(Cq0+shift, temp, nt); 
        j  = locate(Cq+shift,  temp, nt); 
        mag0 = vv[j0] + (Cq0+shift-temp[j0])*(vv[j0+1]-vv[j0])/(temp[j0+1]-temp[j0]); 
        if (phot[id] == 2) {   /* B photometric band */ 
            *bright   = bb[j] + (Cq+shift-temp[j])*(bb[j+1]-bb[j])/(temp[j+1]-temp[j]); 
            *bright   = pow(10.0, -0.4 * (*bright-mag0));  
            *dbright  = db[j] + (Cq+shift-temp[j])*(db[j+1]-db[j])/(temp[j+1]-temp[j]); 
            *dbright *= -0.4 * log(10.0) * *bright; 
        } else if (phot[id] == 3) {   /* V photometric band */ 
            *bright   = vv[j] + (Cq+shift-temp[j])*(vv[j+1]-vv[j])/(temp[j+1]-temp[j]); 
            *bright   = pow(10.0, -0.4 * (*bright-mag0));  
            *dbright  = dv[j] + (Cq+shift-temp[j])*(dv[j+1]-dv[j])/(temp[j+1]-temp[j]); 
            *dbright *= -0.4 * log(10.0) * *bright; 
        }
    }
}


void sample(float *I, float *dI, int deriv, double *lambda, int n, float Cq, float mu, 
            int id) 
{
    int   k, kmu, pix, kbr, kspot; 
    float shift, delta, dbr, tmp1, tmp2, *pt1, *pt2, step; 
    int   locate(float x, float array[], int n); 

    if ((lambda[0] < laml[id]) || (lambda[n-1] > lamu[id])) {
        printf("Tables should span at least [%.4f %.4f]\n", lambda[0], lambda[n-1]); 
        exit(1); 
    } 

    kmu = locate(mu, mmu, nmu); 
    if (2*mu < mmu[kmu]+mmu[kmu+1]) kmu++; 

    kbr = locate(Cq, br, nbr); 
    dbr = (Cq - br[kbr]) / (br[kbr+1]-br[kbr]); 
/*
    if (dbr < 0.01) 
        dbr = 0.0; 
    else if (dbr > 0.99) { 
        kbr++; 
        dbr = 0.0; 
    }
*/
    
/*
printf("%f %d %f, %f %d %f\n", mu, kmu, mmu[kmu], Cq, kbr, br[kbr]); 
*/

    step = (lamu[id]-laml[id]) / (nlam-1); 
    if (deriv == 0) {
        if (dbr != 0.0) {
#pragma _CNX no_recurrence
            for (k = 0; k < n; k++) { 
                shift = (lambda[k]-laml[id])/step; 
                pix = (int) shift; 
                delta = shift-pix; 
                pt1 = prof[id][kmu][kbr]; 
                pt2 = prof[id][kmu][kbr+1]; 
                tmp1 = (1.0-delta) * pt1[pix] + delta * pt1[pix+1]; 
                tmp2 = (1.0-delta) * pt2[pix] + delta * pt2[pix+1]; 
                I[k] = (1.0-dbr) * tmp1 + dbr * tmp2; 
            }
        } else {
#pragma _CNX no_recurrence
            for (k = 0; k < n; k++) { 
                shift = (lambda[k]-laml[id])/step; 
                pix = (int) shift; 
                delta = shift-pix; 
                pt1 = prof[id][kmu][kbr]; 
                I[k] = (1.0-delta) * pt1[pix] + delta * pt1[pix+1]; 
            }
        }
    } else {
        if (dbr != 0.0) {
#pragma _CNX no_recurrence
            for (k = 0; k < n; k++) { 
                shift = (lambda[k]-laml[id])/step; 
                pix = (int) shift; 
                delta = shift-pix; 
                pt1 = prof[id][kmu][kbr]; 
                pt2 = prof[id][kmu][kbr+1]; 
                tmp1 = (1.0-delta) * pt1[pix] + delta * pt1[pix+1]; 
                tmp2 = (1.0-delta) * pt2[pix] + delta * pt2[pix+1]; 
                I[k]  = (1.0-dbr) * tmp1 + dbr * tmp2; 
                pt1 = dprof[id][kmu][kbr]; 
                pt2 = dprof[id][kmu][kbr+1]; 
                tmp1 = (1.0-delta) * pt1[pix] + delta * pt1[pix+1]; 
                tmp2 = (1.0-delta) * pt2[pix] + delta * pt2[pix+1]; 
                dI[k] = (1.0-dbr) * tmp1 + dbr * tmp2; 
            }
        } else {
#pragma _CNX no_recurrence
            for (k = 0; k < n; k++) { 
                shift = (lambda[k]-laml[id])/step; 
                pix = (int) shift; 
                delta = shift-pix; 
                pt1 =  prof[id][kmu][kbr]; 
                pt2 = dprof[id][kmu][kbr]; 
                I[k]  = (1.0-delta) * pt1[pix] + delta * pt1[pix+1]; 
                dI[k] = (1.0-delta) * pt2[pix] + delta * pt2[pix+1]; 
            }
        }
    }
}


int locate(float x, float array[], int n)
{
    int klo, khi, k, ascnd;

    klo = 0;
    khi = n-1;
    ascnd = (array[n-1] > array[0]); 
    while (khi-klo > 1) {
        k = (khi + klo)/2;
        if (x > array[k] == ascnd) 
            klo = k;
        else
            khi = k;
    }
    return klo;
}


int read_table(char filename[], int batch)
{
    int   nl, j, k, l, m, *indx; 
    float tmp, *mult; 
    char  *name[80]; 
    FILE  *fp; 

    printf("Name of file to read Kurucz table from : ");
    scanf("%s", filename);
    if (batch == 1) printf("%s\n",filename);
    
    fp = fopen(filename,"r");
    if (fp == NULL) return 1;

    fscanf(fp, "%d %d %d %d", &nl, &nmu, &nbr, &nlam); 
    if (nmu) mmu  =  vector(0, nmu-1); 
    if (nbr) br   =  vector(0, nbr-1); 
    laml = dvector(0, NUM_LINES-1); 
    lamu = dvector(0, NUM_LINES-1); 
    if (nl)  mult =  vector(0, nl-1); 
    if (nl)  indx = ivector(0, nl-1); 
    if (nmu*nbr*nlam) prof  = matrix4D(0, NUM_LINES-1, 0, nmu-1, 0, nbr-1, 0, nlam-1); 
    if (nmu*nbr*nlam) dprof = matrix4D(0, NUM_LINES-1, 0, nmu-1, 0, nbr-1, 0, nlam-1); 

    for (j = 0; j < nl; j++) { 
        fscanf(fp, "%s", name); 
        indx[j] = -1; 
        for (k = 0; k < NUM_LINES; k++) { 
            if ((!strcmp(name,line_names[k])) && (!mod[k]) && (!phot[k])) { 
                mod[k] = 1; 
                indx[j] = k; 
                continue; 
            } 
        } 
    } 
    for (j = 0; j < NUM_LINES; j++)  
        if ((!mod[j]) && (!phot[j])) 
            sprintf(line_names[j],""); 

    for (j = 0; j < nmu; j++) fscanf(fp, "%f",  mmu+j); 
    for (j = 0; j < nbr; j++) fscanf(fp, "%f",  br+j); 
    for (j = 0; j < nl;  j++) if (indx[j] >= 0) fscanf(fp, "%lf", laml+indx[j]); 
                              else              fscanf(fp, "%*lf");
    for (j = 0; j < nl;  j++) if (indx[j] >= 0) fscanf(fp, "%lf", lamu+indx[j]); 
                              else              fscanf(fp, "%*lf");

    if (!ap) 
        for (j = 0; j < nl; j++) { 
            if (indx[j] >= 0)  { 
                printf("Multiplication factor for line %s : ", line_names[indx[j]]); 
                scanf("%f", mult+j); 
                if (batch) printf("%f\n", mult[j]); 
            } 
        } 
    else 
        for (j = 0; j < nl; j++) mult[j] = 1.0; 
    
    for (j = 0; j < nl; j++)  
        if (indx[j] >= 0) { 
            for (k = 0; k < nmu; k++) 
                for (l = 0; l < nbr; l++) 
                    for (m = 0; m < nlam; m++) { 
                        fscanf(fp, "%f", &tmp); 
                        prof[indx[j]][k][l][m] = 1.0-mult[j]*(1.0-tmp);
                    } 
        } else { 
            for (k = 0; k < nmu; k++) 
                for (l = 0; l < nbr; l++) 
                    for (m = 0; m < nlam; m++)  
                        fscanf(fp, "%*f"); 
        } 
        

    for (j = 0; j < nl; j++)  
        if (indx[j] >= 0) { 
            for (k = 0; k < nmu; k++) 
                for (l = 0; l < nbr; l++) 
                    for (m = 0; m < nlam; m++) {
                        fscanf(fp, "%f", &tmp); 
                        dprof[indx[j]][k][l][m] = tmp*mult[j]; 
                    } 
        } else { 
            for (k = 0; k < nmu; k++) 
                for (l = 0; l < nbr; l++) 
                    for (m = 0; m < nlam; m++)  
                        fscanf(fp, "%*f"); 
        } 
 
    fclose(fp); 
    if (nl) free_vector(mult,  0, nl-1); 
    if (nl) free_ivector(indx, 0, nl-1); 

    return 0; 
}
