/*

  This package implements the maximum entropy image
  reconstruction algorithm presented by
  J.Skilling and R.K.Bryan, (1984) Mon.Not.R.Astr.Soc. 211, 111-124.

  This routine solves for the maximum entropy image for both linear
  and non-linear image reconstruction problems.
  Each call to the routine mem_nl() takes one iteration from the current
  image towards the maximum entropy image.
  The program tries to follow the notation of Skilling and Bryan as
  much as possible.  All page and equation references in the program
  comments refer to their paper.

  The routine mem_nl() is called by the driving program.  It advances the
  solution one iteration.

  Input Parameters:
            n - number of image elements in the image vector I.
           n1 - use +ve only entropy for the first n1 elements.  Use
                modified entropy for the other n - n1 elements.
            m - number of data elements in the data vector D.
            I - array of n image values with subscripts in the range
                0..(n-1).  On input I holds the current image.
                I corresponds to the f of Skilling and Bryan [Eqn 1].
            D - array of m data values the image is to be fitted to
                [Eqn 2].  The subscripts must be in the range 0..(m-1).
            F - for non-linear problems this array must contain m data
                values corresponding to the image in I.  For linear
                problems F is used as a work array.  The subscripts must
                be in the range 0..(m-1).
         sig2 - array of m squared standard deviations of the data values held
                in D.  The subscripts must be in the range 0..(m-1).
            R - response matrix held in a matrix allocated with the
                matrix routine in {mem\_alloc.w}.  The row subscript must be
                in the range 0..(m-1) and the column subscript in the
                range 0..(n-1).  Each element of the response matrix
                is defined as R(k,j) = dF(k)/dI(j).
            A - default intensity [Page 114, Eqn 6].
         chi0 - the desired chi-squared fit statistic [Eqn 4].
      Rchange - used only in linear image reconstruction.  Rchange must
                be set to one on the first call to |mem_nl()| and on
                subsequent calls where the response matrix, image vector or
                data vector have been changed.  Rchange should be set to
                zero if these parameters are unchanged and the code will
                execute a bit faster.
         nlin - must be set to one for non-linear problems, zero otherwise.

  Output Parameters:
            I - updated image vector.
      entropy - for linear problems entropy is set to the current image
                entropy [Eqn 6].  For non-linear problems it contains the
                entropy of the input image.
         chi2 - for linear problems chi2 is set to the current chi-squared
                statistic [Eqn 4]. For non-linear problems it contains the
                chi-squared value at the start of the iteration.
         test - for linear problems test is set to the entropy test
                recommended by Skilling and Bryan [Eqn 37] for the current
                image.  For non-linear problems it contains the test value
                of the initial image.

  The parameter DNE sets the number of search directions used.
*/

#define COMPILING_MEM_NL
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "mem_alloc.h"
#include "mem_nl.h"


#define DNE 6

static double *gS,*gC,*ggS,*aI,C0,S0,Itot;


void mem_nl(n1,n2,n,m,I,D,F,sig2,R,w,A,B,C,Chi0,Rchange,entropy,Chi2,test,nlin,L_fac)
int   n1,n2,n,m;
float I[],D[],F[],sig2[],**R,w[];
float A,B,*C,Chi0;
int   Rchange,nlin;
float *entropy,*Chi2,*test,L_fac;
{
    double Smu[DNE],Cmu[DNE],gamma[DNE],x[DNE],almin,Caim,L0,**e;
    int    ne;

    if (nlin) {
        gS = dvector(0,n-1);
        gC = dvector(0,n-1);
        ggS = dvector(0,n-1);
        aI  = dvector(0,n-1);
    } else {
        if (Rchange == -1) {
            gS = dvector(0,n-1);
            gC = dvector(0,n-1);
            ggS = dvector(0,n-1);
            aI  = dvector(0,n-1);
            return;
        } else if (Rchange == -2) {
            free_dvector(gS,0,n-1);
            free_dvector(gC,0,n-1);
            free_dvector(ggS,0,n-1);
            free_dvector(aI,0,n-1);
            return;
        }
    }

    e = dmatrix(0,n-1,0,DNE-1);

/* If first call or the response matrix R has changed it is necessary to
   calculate the gradient information. */

    if (Rchange || nlin) {
        calc_grad(n1,n2,n,m,I,D,F,sig2,R,w,A,B,C,nlin,ggS,aI);
        calc_test(n,test);
      }

/* Calculate the normalised search directions. */

    ne = DNE;
    search(e,ne,aI,n,m,sig2,R);

/* Now diagonalise the search subspace and construct quadratic models
   for S and C. */

    diagmod(e,gamma,Smu,Cmu,&ne,n,m,ggS,sig2,R);

/* Calculate the constants needed by the CONTROL subroutine. */

    consts(gamma,ne,&almin,&Caim,&L0,Chi0,C0,Cmu,Itot,L_fac);

/* Now calculate the increments x to move in each of the search directions. */

    control(x,S0,C0,Smu,Cmu,gamma,Caim,L0,almin,ne);

/* Update the image vector along the search directions. */

    update(x,e,ne,I,n1,n2,n,A);

/* Calculate the gradient information which is needed to calculate test. */

    if (!nlin) {
        calc_grad(n1,n2,n,m,I,D,F,sig2,R,w,A,B,C,nlin,ggS,aI);
        calc_test(n,test);  /* Calculate test. */
    }
    *entropy = S0;
    *Chi2 = C0;

    if (nlin) {
        free_dvector(gS,0,n-1);
        free_dvector(gC,0,n-1);
        free_dvector(ggS,0,n-1);
        free_dvector(aI,0,n-1);
    }
    free_dmatrix(e,0,n-1,0,DNE-1);
}


static void calc_grad(n1,n2,n,m,I,D,F,sig2,R,w,A,B,C,nlin,ggS,aI)
int    n1,n2,n,m;
float  I[],D[],F[],sig2[],**R,w[];
float  A,B,*C;
int    nlin;
double ggS[],aI[];
{
    double sum1,sum2,la,li,tmp,tmp2,*sumvec,CC,II;
    int    j,k;

/* Finds the simulated data for the current image [Eqn 3] and the
   current chi-squared value [Eqn 4]. */

    if (!nlin) {
        sum2 = 0.0;
        for (k = 0; k < m; k++) {
            sum1 = 0.0;
#pragma _CNX no_recurrence
            for (j = 0; j < n; j++)
                sum1 += R[k][j]*I[j];
            F[k] = sum1;
            tmp  = sum1 - D[k];
            sum2 += tmp*tmp/sig2[k];
        }
        C0 = sum2;
    } else {
        sum2 = 0.0;
        for (k = 0; k < m; k++) {
            tmp = F[k] - D[k];
            sum2 += tmp*tmp/sig2[k];
        }
        C0 = sum2;
    }

/* Finds the current image entropy [Eqn 6], gradient of S [Eqn 7a] and diagonal elements
   of the second derivative matrix ggS (other elements are zero). */

    la = log(A);
    sum1 = Itot = 0.0;
#pragma _CNX no_recurrence
    for (j = 0; j < n1; j++) {
        li = log(I[j]);
        gS[j]  = w[j]*(la - li);
        ggS[j] = w[j]/I[j];       /* Store negative diag(ggS) */
        aI[j] = 1.0/ggS[j];
/*
        sum1 -= w[j]*I[j]*(log(I[j]/A) - 1.0);
*/
        sum1 -= w[j]*(I[j]*(log(I[j]/A) - 1.0) + A);
        Itot += w[j]*I[j];
    }

   CC = (double) 0.999*A;   
#pragma _CNX no_recurrence
    for (j = n1; j < n2; j++) {   /* "filling factor" entropy */
        gS[j]  = w[j]*(log(C[j]/I[j])-log((A-C[j])/(A-I[j])));
        ggS[j] = w[j]*A/(I[j]*(A-I[j]));  /* Negative diag(ggS) */    
	aI[j] = 1.0/ggS[j];
        sum1 -= w[j]*(I[j]*log(I[j]/C[j]) + (A-I[j])*log((A-I[j])/(A-C[j])));
        Itot += w[j];
    }

#pragma _CNX no_recurrence
    for (j = n2; j < n; j++) {
        II = I[j]-C[j];
        tmp = sqrt(II*II + 4.0*B*B);
        gS[j] = -w[j]*log((tmp+II)/(2.0*B));
        ggS[j] = w[j]/tmp;              /* Store negative diag(ggS) */
	aI[j]  = 1.0 / ggS[j];
        tmp2 = fabs(II);
        if (tmp2 < B) tmp2 = B;
        tmp2 *= w[j];
        Itot += tmp2;
        sum1 += w[j]*(tmp - 2.0*B - II*log((tmp+II)/(2.0*B)));
    }
    S0 = sum1;

/* Finds the gradient of C [Eqn 8]. */

    sumvec = dvector(0,n-1);
    for (j = 0; j < n; j++)
        sumvec[j] = 0.0;
    for (k = 0; k < m; k++)
#pragma _CNX no_recurrence
        for (j = 0; j < n; j++)
            sumvec[j] += R[k][j]*(F[k]-D[k])/sig2[k];
    for (j = 0; j < n; j++)
        gC[j] = 2.0*sumvec[j];
    free_dvector(sumvec,0,n-1);
}


static void search(e,ne,aI,n,m,sig2,R)
double **e,aI[];
float  sig2[],**R;
int    ne,n,m;
{
    double *temp,sum1,sum2,tmp,*sumvec;
    int    j,k,p,gceq0;

    temp = dvector(0,m-1);
    sumvec = dvector(0,n-1);

/* Routine finds |DNE| search directions according to Eqn 20 and the
   paragraph above Eqn 19.
   Finds f(gC) The resulting search direction is normalised. */

    gceq0 = 0;
    sum1 = 0.0;
#pragma _CNX no_recurrence
    for (j = 0; j < n; j++) {
        e[j][0] = tmp = gC[j]*aI[j];
        sum1 += tmp*tmp;
    }
    if (sum1 > 0.0)
        sum1 = sqrt(sum1);
    else {
        sum1 = 1.0;
        gceq0 = 1;
    }
    for (j = 0; j < n; j++)
        e[j][0] /= sum1;
    if (ne == 1) return;

/* Finds f(gS) The resulting search direction is normalised. */

    sum1 = 0.0;
#pragma _CNX no_recurrence
    for (j = 0; j < n; j++) {
        e[j][1] = tmp = gS[j]*aI[j];
        sum1 += tmp*tmp;
    }
    if (sum1 > 0.0)
        sum1 = sqrt(sum1);
    else {
        sum1 = 1.0;
        if (gceq0 == 1) {
            fprintf(stderr,"Error: f(gS) and f(gC) are both zero vectors\n");
            exit(1);
        }
    }
    for (j = 0; j < n; j++)
        e[j][1] /= sum1;

/* Loops around multiplying direction p-2 by matrix f(ggC) to get
   direction p which is then normalised. */

    for (p = 2; p < ne; p++) {
        for (k = 0; k < m; k++) {
            sum2 = 0.0;
#pragma _CNX no_recurrence
            for (j = 0; j < n; j++)
                sum2 += R[k][j]*e[j][p-2];
            temp[k] = sum2;
        }
        sum1 = 0.0;
        for (j = 0; j < n; j++)
            sumvec[j] = 0.0;
        for (k = 0; k < m; k++)
#pragma _CNX no_recurrence
            for (j = 0; j < n; j++)
                sumvec[j] += 2.0*aI[j]*R[k][j]*temp[k]/sig2[k];
        for (j = 0; j < n; j++) {
            e[j][p] = tmp = sumvec[j];
            sum1 += tmp*tmp;
        }
        if (sum1 > 0.0)
            sum1 = sqrt(sum1);
        else
            sum1 = 1.0;
        for (j = 0; j < n; j++)
            e[j][p] /= sum1;
    }

    free_dvector(sumvec,0,n-1);
    free_dvector(temp,0,m-1);
}


static void diagmod(e,gamma,Smu,Cmu,ne,n,m,ggS,sig2,R)
double **e,gamma[],Smu[],Cmu[],ggS[];
int    *ne,n,m;
float  sig2[],**R;
{
    double **e2,**temp,sum1,sum2,**g,**MM,**V,min,max,fgam;
    int    j,k,p,q,qq,ne1,deleted_one,pmin;

    ne1  = *ne - 1;
    e2   = dmatrix(0,n-1,0,ne1);
    temp = dmatrix(0,m-1,0,ne1);
    g    = dmatrix(0,ne1,0,ne1);
    MM   = dmatrix(0,ne1,0,ne1);
    V    = dmatrix(0,ne1,0,ne1);

/* Routine constructs and diagonalises the search subspace according to
   the procedures in Eqn 24 and Section 3.7.1.
   Calculate the metric tensor g [Eqn 24b] (modified to use ggS correctly). */

    do {
	deleted_one = 0;

        for (p = 0; p < *ne; p++) {
            for (q = p; q < *ne; q++) {
                sum1 = 0.0;
                for (j = 0; j < n; j++)
                    sum1 += e[j][p]*e[j][q]*ggS[j];
                g[p][q] = g[q][p] = sum1;
            }
        }

/* Diagonalise g. Eigenvectors are columns in the array V */

        jacobi(g,*ne,gamma,V,&qq);

/* If any of the eigenvalues of g are small then throw out that direction
   and repeat the calculation of g etc.  This protects against linear
   dependence in the search directions.  NOTE the factor 1.E-4 is an
   arbitrary choice. */

	min = max = fabs(gamma[0]);  pmin = 0;
        for (p = 1; p < *ne; p++) {
	    fgam = fabs(gamma[p]);
	    if (fgam > max)
		max = fgam;
	    if (fgam < min) {
		min = fgam;
		pmin = p;
	    }
        }

        if (min < 1.0E-4*max) {
	    deleted_one = 1;
            for (p = pmin+1; p < *ne; p++) {
                for(j = 0; j < n; j++)
                    e[j][p-1] = e[j][p];
                gamma[p-1] = gamma[p];
	    }
	    --(*ne);
        }

    } while (deleted_one);

/* Rescale the search directions so that the metric is cartesian.  The
   new directions are held in the matrix e2. */

    for (j = 0; j < n; j++) {
        for (p = 0; p < *ne; p++) {
            sum1 = 0.0;
            for (q = 0; q < *ne; q++)
                sum1 += e[j][q]*V[q][p];
            e2[j][p] = sum1 / sqrt(gamma[p]);
        }
    }

/* Now calculate M (held in matrix MM) [Eqn 24d]. */

    for (k = 0; k < m; k++) {
        for (p = 0; p < *ne; p++) {
            sum1 = 0.0;
            for (j = 0; j < n; j++)
                sum1 += R[k][j]*e2[j][p];
            temp[k][p] = sum1;
        }
    }
    for (p = 0; p < *ne; p++) {
        for (q = p; q < *ne; q++) {
            sum1 = 0.0;
            for (k = 0; k < m; k++)
                sum1 += 2.0*temp[k][p]*temp[k][q]/sig2[k];
            MM[p][q] = MM[q][p] = sum1;
        }
    }

/* Diagonalise M. Eigenvectors are columns in the array V */

    jacobi(MM,*ne,gamma,V,&qq);

/* Rescale the search directions so that M is diagonal.  The
   new directions are held in the matrix e. */

    for (j = 0; j < n; j++) {
        for (p = 0; p < *ne; p++) {
            sum1 = 0.0;
            for (q = 0; q < *ne; q++)
                sum1 += e2[j][q]*V[q][p];
            e[j][p] = sum1;
        }
    }

/* Calculate the final components of the quadratic models Smu and Cmu
   [Eqns 24a and 24c]. */

    for (p = 0; p < *ne; p++) {
        sum1 = sum2 = 0.0;
        for (j = 0; j < n; j++) {
            sum1 += e[j][p]*gS[j];
            sum2 += e[j][p]*gC[j];
        }
        Smu[p] = sum1;
        Cmu[p] = sum2;
    }

    free_dmatrix(e2,  0,n-1,0,ne1);
    free_dmatrix(temp,0,m-1,0,ne1);
    free_dmatrix(g,   0,ne1,0,ne1);
    free_dmatrix(MM,  0,ne1,0,ne1);
    free_dmatrix(V,   0,ne1,0,ne1);
}


static void consts(gamma,ne,almin,Caim,L0,Chi0,C0,Cmu,Itot,L_fac)
int    ne;
double gamma[],*almin,*Caim,*L0,C0,Cmu[],Itot;
float  Chi0,L_fac;
{
    double sum1,Cmin;
    int    p;

/* Routine calculates the constants needed by the control procedure.
   Find Cmin [Eqn 28]. */

    sum1 = 0.0;
    for (p = 0; p < ne; p++)
        sum1 += Cmu[p]*Cmu[p]/gamma[p];
    Cmin = C0 - 0.5*sum1;

/* Find Caim [Eqn 29].  The constants 0.8 and 0.2 are different from
   Skilling and Bryans choice of 0.667 and 0.333. */

    *Caim = 0.667*Cmin + 0.333*C0;
    if (Chi0 > *Caim)
        *Caim = Chi0;

/* Find L0 [Paragraph above Eqn 28].  Choice of 0.3 is arbitrary. */

    *L0 = L_fac*Itot;

/* Find alpha min [Paragraph below Eqn 32]. */

    *almin = -gamma[0];
    for (p = 1; p < ne; p++) {
        sum1 = -gamma[p];
        if (sum1 > *almin) *almin = sum1;
    }
    if (*almin < 0.0) *almin = 0.0;
}


static void control(x,S0,C0,Smu,Cmu,gamma,Caim,L0,almin,ne)
double x[],S0,C0,Smu[],Cmu[],gamma[],Caim,L0,almin;
int    ne;
{
    double alpha,P,C,Cp,L,xs[DNE],alow,Plow,ahigh,Phigh,Caimh,Lsuc,Csuc;
    int    q,asuc,Psuc,afinished,Pfinished;
static int called = 0;
++called;

/* Implements the control procedure charted on Page 122, Figure 3 and
   described in Sections 3.7.2 and 3.7.3. */

    if (C0 < Caim*1.01)
        Caimh = Caim*1.01;
    else
        Caimh = Caim;

    P = Plow = 0.0;
    Phigh = -1.0;
    Psuc = Pfinished = 0;

    while (Pfinished == 0) {                /* Start of the P chop. */

        alow = almin;
        ahigh = -1.0;
        alpha = almin + 1.0;
        asuc = afinished = 0;

        while (afinished == 0) {            /* Start of the alpha chop */

/* Calculate C [Eqn 23b], Cp [Eqn 35], L [Eqn 27c] and x [Eqn 34]. */

            C = Cp = C0;
            L = 0.0;
            for (q = 0; q < ne; q++) {
                x[q] = (alpha*Smu[q] - Cmu[q]) / (P+gamma[q]+alpha);
                C   += Cmu[q]*x[q] + 0.5*gamma[q]*x[q]*x[q];
                Cp  += Cmu[q]*x[q] + 0.5*(P+gamma[q])*x[q]*x[q];
                L   += x[q]*x[q];
            }

            if ((Cp > C0) && (Cp > Caimh)) {
                ahigh = alpha;
                alpha = 0.5*(alow + alpha);
            } else if (L > L0) {
                if (Cp <= C0) {
                    alow = alpha;
                    if (ahigh > 0.0)
                        alpha = 0.5*(ahigh + alpha);
                    else
                        alpha = 2.0*alpha;
                } else {
                    ahigh = alpha;
                    alpha = 0.5*(alow + alpha);
                }
            } else if (C < Caim) {
	       /* Commented code below was in an old version */
	       /* It is only useful if the initial chi^2 < initial Caim */
               /* if (Caimh >= C0) {
                    for (q = 0; q < ne; q++)
                        xs[q] = x[q];
                    asuc = 1;
                    Lsuc = L;
                    Csuc = C;
                } */
                alow = alpha;
                if (ahigh > 0.0)
                    alpha = 0.5*(ahigh + alpha);
                else
                    alpha = 2.0*alpha;
            } else {
                for (q = 0; q < ne; q++)
                    xs[q] = x[q];
                asuc = 1;
                Lsuc = L;
                Csuc = C;
                ahigh = alpha;
                alpha = 0.5*(alow + alpha);
            }
            if ((ahigh > 0.0) && ((ahigh-alow) < (1.0E-3*ahigh+1.0E-10)))
                afinished = 1;
            if (alpha > 1.0E20)
                afinished = 1;
        }                                   /* End of alpha chop */

        if (asuc == 1) {
            Psuc = 1;
            Phigh = P;
            P = 0.5*(Plow + P);
        } else {
            Plow = P;
            if (Phigh > 0.0)
                P = 0.5*(Phigh + P);
            else
                P = 2.0*(P + 1.0);
        }

        if ((Psuc == 1) && ((P == 0.0) || (Phigh-Plow) < (1.0E-3*Phigh+1.0E-10)))
            Pfinished = 1;

        if ((Psuc == 0) && (P > 1.0E20)) {  /* Check for occasional   */
            Pfinished = 1;                  /* bad topology. (rare)   */ 
            printf("P chop blow up encountered.\n");
        }
    }                                       /* End of P chop */


/* printf("ne = %d, C0 = %9.4e, Caim = %9.4e, alpha = %9.4e, P = %9.4e\n",ne,C0,Caim,alpha,P);
   printf("Caimh = %9.4e, L = %9.4e, L0 = %9.4e, almin = %9.4e, C = %9.4e\n",
                                                                 Caimh,Lsuc,L0,almin,Csuc); */ 


    if (Psuc == 0)                          /*  Topology Correction  */
        for (q = 0; q < ne; q++)
            x[q] = 1.0e-2*L0*(Smu[q]-Cmu[q]);

/* Copy the most recent successful x (array xs) to the output increment
   array x. */

    for (q = 0; q < ne; q++) {
        x[q] = xs[q];
        /*printf("x[%1d] = %f, ",q+1,x[q]);*/
    }
    /*printf("\n");*/
}


static void update(x,e,ne,I,n1,n2,n,A)
double x[],**e;
float  I[],A;
int    ne,n1,n2,n;
{
    int j, p;

/* Update the image vector [Eqn 25]. */

    for (p = 0; p < ne; p++) {
#pragma _CNX no_recurrence
        for (j = 0; j < n; j++)
            I[j] += x[p]*e[j][p];
    }

/* Protect against stray negative image values. */

    for (j = 0; j < n1; j++)
        if (I[j] <= 0.0)
            I[j] = 1.0E-6;

    for (j = n1; j < n2; j++) {
        if (I[j] <= 0.0)
            I[j] = 1.0E-6*A;
        if (I[j] >= A)
            I[j] = A*(1.0-1.0E-6);
    }
}


static void calc_test(n,test)
int n;
float *test;
{
    double sum1,sum2,sum3,tmp;
    int    j;

/* Finds Skilling and Bryans test parameter [Eqn 37]. */

    sum1 = sum2 = 0.0;
    for (j = 0; j < n; j++) {
        sum1 += gS[j]*gS[j];
        sum2 += gC[j]*gC[j];
    }
    sum1 = sqrt(sum1);
    sum2 = sqrt(sum2);
    sum3 = 0.0;
    if (sum1 != 0.0 && sum2 != 0.0) {
        for (j = 0; j < n; j++) {
            tmp = gS[j]/sum1 - gC[j]/sum2;
            sum3 += tmp*tmp;
        }
    }
    *test = 0.5*sum3;
}


static void jacobi(a,n,d,v,nrot)
double **a,d[],**v;
int    n,*nrot;
{
    double *b,*z,sm,tresh,g,h,t,theta,c,s,tau;
    int    ip,iq,i,j;

    b = dvector(0,n-1);
    z = dvector(0,n-1);

    for (ip = 1; ip <= n; ip++) {
        for (iq=1; iq <= n; iq++)
            *(*(v+ip-1)+iq-1) = 0.0;
        *(*(v+ip-1)+ip-1) = 1.0;
    }
    for (ip = 1; ip <= n; ip++) {
        b[ip-1] = *(*(a+ip-1)+ip-1);
        d[ip-1] = b[ip-1];
        z[ip-1] = 0.0;
    }
    *nrot = 0;
    for (i = 1; i <= 50; i++) {
        sm = 0.0;
        for (ip = 1; ip <= n-1; ip++) {
            for (iq = ip+1; iq <= n; iq++)
                sm += fabs(*(*(a+ip-1)+iq-1));
        }
        if (sm == 0.0) {
            free_dvector(b,0,n-1);
            free_dvector(z,0,n-1);
            return;
        }
        if (i < 4)
            tresh = 0.2*sm/(n*n);
        else
            tresh = 0.0;
        for (ip = 1; ip <= n-1; ip++) {
            for (iq=ip+1; iq <= n; iq++) {
                g = 100.0*fabs(*(*(a+ip-1)+iq-1));
                if (i > 4 && fabs(d[ip-1])+g == fabs(d[ip-1])
                    && fabs(d[iq-1])+g == fabs(d[iq-1]))
                    *(*(a+ip-1)+iq-1) = 0.0;
                else if (fabs(*(*(a+ip-1)+iq-1)) > tresh) {
                    h = d[iq-1]-d[ip-1];
                    if (fabs(h)+g == fabs(h))
                        t = *(*(a+ip-1)+iq-1)/h;
                    else  {
                        theta = 0.5*h/(*(*(a+ip-1)+iq-1));
                        t = 1.0/(fabs(theta)+sqrt(1.0+theta*theta));
                        if (theta < 0.0) t = -t;
                    }
                    c = 1.0/sqrt(1 + t*t);
                    s = t*c;
                    tau = s/(1.0+c);
                    h = t*(*(*(a+ip-1)+iq-1));
                    z[ip-1] -= h;
                    z[iq-1] += h;
                    d[ip-1] -= h;
                    d[iq-1] += h;
                    *(*(a+ip-1)+iq-1) = 0.0;
                    for (j = 1; j <= ip-1; j++) {
                        g = *(*(a+j-1)+ip-1);
                        h = *(*(a+j-1)+iq-1);
                        *(*(a+j-1)+ip-1) = g-s*(h+g*tau);
                        *(*(a+j-1)+iq-1) = h+s*(g-h*tau);
                    }
                    for (j = ip+1; j <= iq-1; j++) {
                        g = *(*(a+ip-1)+j-1);
                        h = *(*(a+j-1)+iq-1);
                        *(*(a+ip-1)+j-1) = g-s*(h+g*tau);
                        *(*(a+j-1)+iq-1) = h+s*(g-h*tau);
                    }
                    for (j = iq+1; j <= n; j++) {
                        g = *(*(a+ip-1)+j-1);
                        h = *(*(a+iq-1)+j-1);
                        *(*(a+ip-1)+j-1) = g-s*(h+g*tau);
                        *(*(a+iq-1)+j-1) = h+s*(g-h*tau);
                    }
                    for (j = 1; j <= n; j++) {
                        g = *(*(v+j-1)+ip-1);
                        h = *(*(v+j-1)+iq-1);
                        *(*(v+j-1)+ip-1) = g-s*(h+g*tau);
                        *(*(v+j-1)+iq-1) = h+s*(g-h*tau);
                    }
                    (*nrot)++;
                }
            }
        }
        for (ip = 1; ip <= n; ip++) {
            b[ip-1] += z[ip-1];
            d[ip-1] = b[ip-1];
            z[ip-1] = 0.0;
        }
    }
    printf("50 iterations should never happen\n");
}
