#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "line.h"
#include "structures.h"

static float Cq0; 
static float Cq1; 

/*  We use static variables to hold the line information.  */

#define NUM_LINES 3
static char *line_names[NUM_LINES] = { "Mean",
		                       "Ca_650",
		                       "phot" };
static int line_ids[NUM_LINES] = {0,1,2};
static int line_lin[NUM_LINES] = {1,1,1};

static double l0[NUM_LINES]  = {          /* $\lambda_0$ */
                                 550.0000,
			 	 650.0000,
			 	 550.0000 }; 
static double dl[NUM_LINES]  = {          /* $\Delta \lambda$ */
				 75.0E-4,
				 70.0E-4,
				 100.0   }; 
static double eps[NUM_LINES] = {         /* $\epsilon$ */
                                 0.75,
				 0.70, 
				 0.70 }; 
static double str[NUM_LINES] = {         /* Line ``strength'' */
                                 0.51, 
				 0.70 }; 
static double g[NUM_LINES] =   {         /* Lande g factor */
                                 1.2,
				 0.5 }; 
static double zeta_rt[NUM_LINES] = {     /* $\zeta$ */
                                 0.00,
				 0.00, 
				 0.00 }; 


void init_spectral_lines(batch)
int batch;
{
/*
    printf("What is the doppler width for the Fe line : ");
    scanf("%lf",&dl[0]);
    if (batch) printf("%g\n",dl[0]);

    printf("What is the line strength for the Fe line : ");
    scanf("%lf",&str[0]);
    if (batch) printf("%g\n",str[0]);

    printf("What is the limb darkening coefficient for the Fe line : ");
    scanf("%lf",&eps[0]);
    if (batch) printf("%g\n",eps[0]);
*/
}


void end_spectral_lines()
{
}


void get_spectral_lines(nlines,names,ids)
int *nlines,**ids;
char **names[];
{
    *nlines = NUM_LINES;
    *names  = line_names;
    *ids    = line_ids;
}


/*
The following routine lets the main program find out some details about
each of the lines.
*/

void line_info(id,ll0,dll,broadened,linear)
int id;      /* Line ID number */
double *ll0;  /* Line centre wavelength */
double *dll;  /* Doppler width */
int *broadened;  /* Does the line model include the instrumental broadening */
int *linear;    /* Is the spectral line model linear */
{
    *ll0 = l0[id];
    *dll = dl[id];
    *broadened = 0;
    *linear = line_lin[id];
}


/*
The function line_centre() needs to return the centre wavelength of the
line being studied.
*/

double find_line_centre(id)
int id;
{
    return l0[id];
}


/*
The function find_macro_turb() needs to return the radial-tangential
macroturbulence parameter.  If you don't need to use macroturbulence
just return zero.  If you do need it you should look at the routines
doppler_spec() and doppler_deriv() to check how it is being
implimented.
*/

double find_macro_turb(id)
int id;
{
    return zeta_rt[id];
}


/*
This routine is used when a spectrum has to be calculated from an image.
You only need to calculate the Stokes parameters that you will use.  For
example, in straight Doppler imaging you only have to calculate Stokes I
using the continuum brightness C_q.
*/

void line_profile(id,I,Q,U,V,n,lambda,Cq,Cm,mu,B,gam,chi,magnetic)
int id;                /* Line ID number */
float I[],Q[],U[],V[];  /* Returned Stokes parameters */
int n;                /* Number of wavelengths to calculate */
double lambda[];        /* Wavelength $\lambda$ */
float Cq,Cm;          /* Continuum brightness */
float mu;              /* Cosine of the limb angle */
float B,gam,chi;     /* Magnetic field strength and orientation $(B,\gamma,\chi)$ */
int magnetic;          /* Switch is non-zero for a magnetic star */
{
    int    j;
    double contin,v0,gg,dIdl,arg,d2Idl2,fac;

    contin = 1.0-eps[id]+eps[id]*mu;

    if (id < 2)  /* Spectral line */
        if (!magnetic) {
#pragma _CNX no_recurrence
            for (j = 0; j < n; j++) {
                v0 = (lambda[j]-l0[id])/dl[id];
	        I[j] = Cq*contin*(1.0 - str[id]*exp(-v0*v0));
            }
        } else {
#pragma _CNX no_recurrence
            for (j = 0; j < n; j++) {
                v0 = (lambda[j]-l0[id])/dl[id];
                arg = exp(-v0*v0); 
	        I[j] = Cq*contin*(1.0 - str[id]*arg);
	        gg = 4.67E-12*lambda[j]*lambda[j]*g[id];
	        dIdl = -2.0*Cq*contin*str[id]*arg*v0/dl[id];
	        V[j] = gg*B*cos(gam)*dIdl;

                d2Idl2 = -20.0*Cq*contin*str[id]*arg*(1.0-2.0*v0*v0)/dl[id]/dl[id];
                fac = gg*B*sin(gam);
                Q[j] = 0.25*fac*fac*cos(2.0*chi)*d2Idl2;
                U[j] = 0.25*fac*fac*sin(2.0*chi)*d2Idl2;
            }
        }
    else  /* Photometry */
	I[0] = Cq*contin;
}


/*
This routine is an extended version of line_profile() used during the
image inversion process.  As well as returning the spectral profile you
also need to return the derivatives of the field profile with respect
to the image quantities you are fitting.
*/

void line_deriv(id,dII,dQQ,dUU,dVV,I,Q,U,V,n,lambda,Cq,Cm,mu,B,gam,chi,magnetic)
int id;
float **dII,**dQQ,**dUU,**dVV;  /* Arrays of profile derivatives */
float I[],Q[],U[],V[];  /* Returned Stokes parameters */
int   n;          /* Number of wavelength points */
double lambda[];    /* Wavelength $\lambda$ */
float Cq,Cm;        /* Continuum brightness */
float mu;           /* Cosine of the limb angle */
float B,gam,chi;    /* Magnetic field strength and orientation $(B,\gamma,\chi)$ */
int magnetic;       /* Switch is non-zero for a magnetic star */
{
    int    j;
    double contin,v0,gg,dIdl,arg,d2Idl2,fac;

    contin = 1.0-eps[id]+eps[id]*mu;

    if (id < 2)  /* Spectral line */
        if (!magnetic) {
#pragma _CNX no_recurrence
            for (j = 0; j < n; j++) {
                v0 = (lambda[j]-l0[id])/dl[id];
	        I[j] = Cq*contin*(1.0 - str[id]*exp(-v0*v0));
	        dII[0][j] = I[j] / Cq;
            }
        } else {
#pragma _CNX no_recurrence
            for (j = 0; j < n; j++) {
                v0 = (lambda[j]-l0[id])/dl[id];
                arg = exp(-v0*v0); 
	        I[j] = Cq*contin*(1.0 - str[id]*arg);
	        dII[0][j] = I[j]/Cq;
	        dII[1][j] = dII[2][j] = dII[3][j] = dII[4][j] = 0.0;
	        gg = 4.67E-12*lambda[j]*lambda[j]*g[id];
	        dIdl = -2.0*Cq*contin*str[id]*arg*v0/dl[id];
	        V[j] = gg*B*cos(gam)*dIdl;
	        dVV[0][j] = V[j] / Cq;
	        dVV[2][j] = gg*cos(gam)*dIdl;
	        dVV[3][j] = gg*B*dIdl;
	        dVV[1][j] = dVV[4][j] = 0.0;

                d2Idl2 = -20.0*Cq*contin*str[id]*arg*(1.0-2.0*v0*v0)/dl[id]/dl[id];
                fac = gg*B*sin(gam);
                Q[j] = 0.25*fac*fac*cos(2.0*chi)*d2Idl2;
                U[j] = 0.25*fac*fac*sin(2.0*chi)*d2Idl2;

                dQQ[0][j] = Q[j]/Cq;
                dQQ[2][j] = 0.5*fac*cos(2.0*chi)*d2Idl2*gg*sin(gam);
                dQQ[3][j] = -0.5*cos(2.0*chi)*d2Idl2*gg*gg*B*B*cos(gam);
                dQQ[4][j] = -2.0*U[j];
                dUU[0][j] = U[j]/Cq;
                dUU[2][j] = 0.5*fac*sin(2.0*chi)*d2Idl2*gg*sin(gam);
                dUU[3][j] = -0.5*sin(2.0*chi)*d2Idl2*gg*gg*B*B*cos(gam);
                dUU[4][j] = 2.0*Q[j];
                dQQ[1][j] = dUU[1][j] = 0.0;
            }
        }
    else {  /* Photometry */
	I[0] = Cq*contin;
	dII[0][0] = contin;
    }
}


void continuum(id,mu,Cq,Cm,cont,dCq,dCm,magnetic)
int id,magnetic;
float mu,Cq,Cm,*cont,*dCq,*dCm;
{
    if (!magnetic) {
        *dCq = 1.0-eps[id]+eps[id]*mu;
        *cont = Cq*(*dCq);
    } else {
        *dCq = 1.0-eps[id]+eps[id]*mu;
        *cont = Cq*(*dCq);
    }
}


void image_info(Cm_used,qnames,ff_entropy,ff_rec,batch)
int *Cm_used,*ff_entropy,*ff_rec, batch;
char *qnames[];
{
    static char *bright = "brightness";

    *Cm_used = 0;
    *ff_entropy  = 0;
    *ff_rec      = 0;
    qnames[0] = bright;
/*
    Cq0 = dCq; 
    Cq1 = dCq_spot; 
*/
}


void transform(Cq,Cm,ngrid,magnetic)
float Cq[],Cm[];
int ngrid,magnetic;
{
}


void untransform(Cq,Cm,ngrid,magnetic)
float Cq[],Cm[];
int ngrid,magnetic;
{
}

void transform_photometry(spec_info, nobs, I, mag0, wphot)
spec_info_t *spec_info;
int nobs;
float *I, mag0, wphot;
{ 
}

void untransform_photometry(spec_info, nobs, I, mag0, wphot)
spec_info_t *spec_info;
int nobs;
float *I, mag0, wphot;
{
} 

void image_defaults(dCq, dB, magnetic, Cm_used, qnames, batch)
float *dCq, *dB;
int magnetic, Cm_used, batch;
char *qnames[];
{
    printf("Input default %s : ",qnames[0]);
    scanf("%f", dCq);
    if (batch) printf("%f\n",*dCq);

    if (magnetic && Cm_used) {
        printf("What is the default %s : ",qnames[1]);
        scanf("%f",dB);
        if (batch) printf("%f\n",*dB);
    }
} 

