/*

This package will broaden a spectral line profile by a Gaussian profile,
for example the instrumental profile or macroturbulent broadening.
*/

#define COMPILING_SMEAR
#include <math.h>
#include <stdio.h>
#include "mem_alloc.h"
#include "smear.h"

#define  TWO_PI  6.283185308       /* $2\pi$  */
#define  PI_sq   9.869604404       /* $\pi^2$ */
#define  LOGe2   0.693147181     /* $\log_e 2$ */

void smear(width,nspec,data,dlambda)
float width;   /* Instrumental width (FWHM), $w$ */
int nspec;     /* Number of points in the line profile */
float data[];  /* The spectral profile, returned with instrumental smearing */
float dlambda; /* The wavelength spacing of the data points, $\Delta\lambda$.  Must have the
                  the same units as |width| */
{
    double *fft_vec,arg,j_to_f_sq,tmp;
    int    fft_size,j;

    if ((width < 0.01*dlambda) || (nspec == 1)) 
        return;   /* No need to do the convolution when $w \ll \Delta\lambda$ */

    fft_size = 1;
    while (fft_size < nspec)
        fft_size *= 2;    /* Find $N$, the first power of 2 $\geq$ |nspec| */
    fft_vec = dvector(0,fft_size-1);

    for (j = 0; j < nspec; j++)
        fft_vec[j] = data[j];  /* Fill the transform array */
    for (j = nspec; j < fft_size; j++)
        fft_vec[j] = data[nspec-1]; /* Pad to power of 2 */

    real_fft(fft_vec,fft_size/2,1);
    arg = -0.25*width*width*PI_sq/LOGe2; /* $-(\beta\pi)^2$ */
    j_to_f_sq = 0.25/(dlambda*dlambda*fft_size*fft_size); /* Converts $j^2$ to $f^2$ */

#pragma _CNX no_recurrence
    for (j = 2; j < fft_size; j += 2) {  /* $f = {j \over 2 N \Delta\lambda}$ */
        tmp = exp(arg*j_to_f_sq*(j*j));    /* $e^{-(\beta\pi f)^2}$ */
        fft_vec[j] *= tmp;
        fft_vec[j+1] *= tmp;
    }
    tmp = exp(arg*j_to_f_sq*fft_size*fft_size); 
    fft_vec[1] *= tmp;          /* The $N^{\rm th}$ element is stored in
                                                       |fft_vec[1]| */

    real_fft(fft_vec,fft_size/2,-1);  /* Now do the inverse transform */
    arg = 2.0 / fft_size;
    for (j = 0; j < nspec; j++)
        data[j] = fft_vec[j] * arg;   /* Normalise it */
    free_dvector(fft_vec,0,fft_size-1);
}


/*
This routine was translated from the Fortran one given
in Numerical Recipes.
*/

static void real_fft(data,n,direction)
double data[];
int n,direction;
{
    double wr,wi,wpr,wpi,wtemp,theta,c1,c2,h1r,h1i,h2r,h2i;
    int    i,i1,i2,i3,i4,n2p3;
#ifdef CONVEX_VECLIB_FFT
    double *work,tmpr,tmpi;
    void z1dfft_();
    long l,iopt,ier;
    int  k;
#endif

#ifdef CONVEX_VECLIB_FFT
    l = n;
    work = dvector(0,5*n/2-1);
    iopt  = -3;
    z1dfft_(data,&l,work,&iopt,&ier);
#endif

    theta = TWO_PI / (2.0*n);
    c1 = 0.5;
    if (direction == 1) {
        c2 = -0.5;
#ifdef CONVEX_VECLIB_FFT
        iopt  = 1;
        z1dfft_(data,&l,work,&iopt,&ier);
	for (k = 2; k < n; k+=2) {
	    tmpr = data[k];         tmpi = data[k+1];
	    data[k] = data[2*n-k];  data[k+1] = data[2*n-k+1];
	    data[2*n-k] = tmpr;     data[2*n-k+1] = tmpi;
        }
#else
        cmpx_fft(data,n,1);
#endif
    } else {
        c2 = 0.5;
        theta = -theta;
    }
    wpr = sin(0.5*theta);
    wpr = -2.0*wpr*wpr;
    wpi = sin(theta);
    wr = 1.0 + wpr;
    wi = wpi;
    n2p3 = 2*n+3;
#pragma _CNX no_recurrence
    for (i = 2; i <= n/2+1; i++) {
        i1 = 2*i - 2;
        i2 = i1 + 1;
        i3 = n2p3 - i2 - 2;
        i4 = i3 + 1;
        h1r = c1*(data[i1] + data[i3]);
        h1i = c1*(data[i2] - data[i4]);
        h2r = -c2*(data[i2] + data[i4]);
        h2i = c2*(data[i1] - data[i3]);
        data[i1] = h1r + wr*h2r - wi*h2i;
        data[i2] = h1i + wr*h2i + wi*h2r;
        data[i3] = h1r - wr*h2r + wi*h2i;
        data[i4] = -h1i + wr*h2i + wi*h2r;
        wtemp = wr;
        wr += wr*wpr - wi*wpi;
        wi += wi*wpr + wtemp*wpi;
    }
    if (direction == 1) {
        h1r = data[0];
        data[0] = h1r + data[1];
        data[1] = h1r - data[1];
    } else {
        h1r = data[0];
        data[0] = c1*(h1r + data[1]);
        data[1] = c1*(h1r - data[1]);
#ifdef CONVEX_VECLIB_FFT
	for (k = 2; k < n; k+=2) {
	    tmpr = data[k];         tmpi = data[k+1];
	    data[k] = data[2*n-k];  data[k+1] = data[2*n-k+1];
	    data[2*n-k] = tmpr;     data[2*n-k+1] = tmpi;
        }
        iopt  = -2;
        z1dfft_(data,&l,work,&iopt,&ier);
#else
        cmpx_fft(data,n,-1);
#endif
    }

#ifdef CONVEX_VECLIB_FFT
    free_dvector(work,0,5*n/2-1);
#endif
}


/*
This routine was translated from the routine four1() in Fortran
given in Numerical Recipes.
*/

static void cmpx_fft(data,nn,direction)
double data[];
int nn,direction;
{
    double wr,wi,wpr,wpi,wtemp,theta,tempr,tempi;
    int    i,istep,j,m,mmax,n;

    n = 2*nn;
    j = 0;
    for (i = 0; i < n; i += 2) {
        if (j > i) {
            tempr = data[j];
            tempi = data[j+1];
            data[j] = data[i];
            data[j+1] = data[i+1];
            data[i] = tempr;
            data[i+1] = tempi;
        }
        m = nn;
        while ((m >= 2) && ((j+1) > m)) {
            j -= m;
            m /= 2;
        }
        j += m;
    }
    mmax = 2;
    while (n > mmax) {
        istep = 2*mmax;
        theta = TWO_PI/(direction*mmax);
        wpr = sin(0.5*theta);
        wpr = -2.0*wpr*wpr;
        wpi = sin(theta);
        wr = 1.0;
        wi = 0.0;
        for (m = 1; m <= mmax; m += 2) {
            for (i = (m-1); i < n; i += istep) {
                j = i+mmax;
                tempr = wr*data[j] - wi*data[j+1];
                tempi = wr*data[j+1] + wi*data[j];
                data[j] = data[i] - tempr;
                data[j+1] = data[i+1] - tempi;
                data[i] = data[i] + tempr;
                data[i+1] = data[i+1] + tempi;
            }
            wtemp = wr;
            wr += wr*wpr - wi*wpi;
            wi += wi*wpr + wtemp*wpi;
        }
        mmax = istep;
    }
}

