/*

A set of routines to perform a variety of ``house-keeping'' tasks that
involve manipulating the selection and management of the observed
spectra.

The routine set_phase_sampling() sets the nsample element of the
spec_info array. This element corresponds to the number of times that the
calculated spectrum must be sampled to ensure that the phase smearing is
less than the value del_ph. It is simply set equal to the number of
times that del_ph divides the observed phase interval, (phasef-phasec),
plus one.
*/

#include "structures.h"
#include "spec_util.h"

void set_phase_sampling(nobs,spec_info,del_ph)
int   nobs;
spec_info_t spec_info[];
float del_ph;
{
    int j;

    for (j = 0; j < nobs; j++)
	spec_info[j].nsample = 1 + (spec_info[j].phasef - spec_info[j].phasec) / del_ph;
}


/*
The routine get_spec_dims() examines the records of the spec_info structure
to determine the total amount of spectral data points and the Stokes
components present in the data.
*/

void get_spec_dims(spec_info,nobs,npns,datsw)
spec_info_t spec_info[];
int nobs,*npns,datsw[];
{
    int j,k;

    for (k = 0; k < 4; k++)
	datsw[k] = 0;

    *npns = 0;

    for (j = 0; j < nobs; j++) {
	*npns += spec_info[j].nspec;
	for (k = 0; k < 4; k++)
	    if (spec_info[j].datsw[k]) datsw[k] = 1;
    }
}


/*
The routine get_selected_stokes() examines the records of the the
spec_info structure to determine the Stokes components that are present in
this selected data set.
*/

void get_selected_stokes(spec_info,nobs,datsw)
spec_info_t spec_info[];
int nobs,datsw[];
{
    int j,k;

    for (k = 0; k < 4; k++)
	datsw[k] = 0;

    for (j = 0; j < nobs; j++)
	if (spec_info[j].selected) {
	    for (k = 0; k < 4; k++)
	        if (spec_info[j].datsw[k]) datsw[k] = 1;
        }
}


/*
This routine finds the number of data points that are selected for fitting.
The result is returned in the variable ntot.
*/

void get_fitted_spec_dims(spec_info,nobs,ntot)
spec_info_t spec_info[];
int nobs,*ntot;
{
    int j,k;

    *ntot = 0;
    for (j = 0; j < nobs; j++)
	if (spec_info[j].selected)
	    for (k = 0; k < 4; k++)
		if (spec_info[j].datsel[k])
		    *ntot += spec_info[j].nspec;
}


/*
Here we check if any of the observations being fitted are spectroscopic
observations (i.e. not photometrically corrected) and, if so, what size
the vector to hold the old continuum derivatives must be.
*/

void check_spectroscopic(spec_info,nobs,imsw,ngrid,spectroscopic,nocd)
spec_info_t spec_info[];
int nobs,imsw[],ngrid,*spectroscopic,*nocd;
{
    int j,ii,n;

    *spectroscopic = *nocd = 0;
    for (j = 0; j < nobs; j++)
	if (spec_info[j].selected)
	    if (!spec_info[j].photometric) {
		*spectroscopic = 1;
		(*nocd)++;
            }

    if (*spectroscopic) {
	n = 0;
	for (ii = 0; ii < 2; ii++)
	    if (imsw[ii])
	        n++;
	*nocd *= n*ngrid;
    }
}


/*
Here we check if the spectral line models of the fitted lines are linear
or not.
*/

void check_linear(spec_info,nobs,linear,quasi_linear)
spec_info_t spec_info[];
int nobs,*linear,*quasi_linear;
{
    int j,ii,n;

    *linear = 1;  *quasi_linear = 0;
    for (j = 0; j < nobs; j++)
	if (spec_info[j].selected) {
	    if (!spec_info[j].linear)
		*linear = 0;
	    if (spec_info[j].linear == 2)
		*quasi_linear = 1;
        }
}
