/*
This package contains routines to build up a spectrum of the star at a
particular phase and over a range of phases.  The routine spectrum()
calculates the spectra one phase and the routine spectra() calculates the
observed spectrum at all phases.
*/

#include <math.h>
#define COMPILING_SPECTRUM
#include "structures.h"
#include "constants.h"
#include "grid.h"
#include "line.h"
#include "mem_alloc.h"
#include "spectrum.h"
#include "smear.h"


void spectra(I,Q,U,V,nobs,spec_info,datsw,sampling,Cq,Cm,Br,Bt,Bp,ngrid,magnetic)
float       I[],Q[],U[],V[];
int         nobs;
spec_info_t spec_info[];
spec_sample_t sampling;
int         datsw[],ngrid,magnetic;
float       Cq[],Cm[],Br[],Bt[],Bp[];
{
    int   i,j,k,l,npns,ns_max;
    float *II,*QQ,*UU,*VV,*B,*gamma,*chi,area,phase,dph,weight;
    static float dummy1=0.0;
    cell_info_t  *cell;
    cell_deriv_t *dummy2;

    npns = ns_max = 0;
    for (j = 0; j < nobs; j++) {
        npns += spec_info[j].nspec;
        if (spec_info[j].nspec > ns_max)
            ns_max = spec_info[j].nspec;
    }

    II = vector(0,ns_max-1);
    QQ = vector(0,ns_max-1);
    UU = vector(0,ns_max-1);
    VV = vector(0,ns_max-1);
    B     = vector(0,ngrid-1);
    gamma = vector(0,ngrid-1);
    chi   = vector(0,ngrid-1);
    cell   = (cell_info_t *) svector(ngrid,(int) sizeof(cell_info_t));
    dummy2 = (cell_deriv_t *) svector(ngrid,(int) sizeof(cell_deriv_t));

    if (datsw[0])  for (k = 0; k < npns; k++)  I[k] = 0.0;
    if (datsw[1])  for (k = 0; k < npns; k++)  Q[k] = 0.0;
    if (datsw[2])  for (k = 0; k < npns; k++)  U[k] = 0.0;
    if (datsw[3])  for (k = 0; k < npns; k++)  V[k] = 0.0;

    npns = 0;
    for (j = 0; j < nobs; j++) {
        dph = (spec_info[j].phasef - spec_info[j].phasec)/(spec_info[j].nsample+1);
	weight = 1.0 / (float) spec_info[j].nsample;
	for (l = 1; l <= spec_info[j].nsample; l++) {
	    phase = spec_info[j].phasec + l*dph;
            geometry(phase,ngrid,cell,magnetic,Br,Bt,Bp,B,gamma,chi,dummy2);
            for (i = 0; i < ngrid; i++) {
                area = cell[i].parea*weight;
                if (area > 0.0) {

		    if (magnetic)
                        doppl_spec(II,QQ,UU,VV,spec_info[j],datsw,sampling,
                                Cq[i],Cm[i],B[i],gamma[i],chi[i],
				cell[i].mu,cell[i].vel,magnetic);
                    else
                        doppl_spec(II,QQ,UU,VV,spec_info[j],datsw,sampling,
                                Cq[i],dummy1,dummy1,dummy1,dummy1,
				cell[i].mu,cell[i].vel,magnetic);

                    if (spec_info[j].datsw[0])
                        for (k = 0; k < spec_info[j].nspec; k++)
                            I[npns+k] += area*II[k];
                    if (spec_info[j].datsw[1])
                        for (k = 0; k < spec_info[j].nspec; k++)
                            Q[npns+k] += area*QQ[k];
                    if (spec_info[j].datsw[2])
                        for (k = 0; k < spec_info[j].nspec; k++)
                            U[npns+k] += area*UU[k];
                    if (spec_info[j].datsw[3])
                        for (k = 0; k < spec_info[j].nspec; k++)
                            V[npns+k] += area*VV[k];
                }
            }
	}
        npns += spec_info[j].nspec;
    }

    free_vector(II,   0,ns_max-1);
    free_vector(QQ,   0,ns_max-1);
    free_vector(UU,   0,ns_max-1);
    free_vector(VV,   0,ns_max-1);
    free_vector(B,    0,ngrid-1);
    free_vector(gamma,0,ngrid-1);
    free_vector(chi,  0,ngrid-1);
    free_svector(cell);
    free_svector(dummy2);
}


/*
The routine spectrum() calculates the spectal intensity between the
wavelength limits laml and lamu, sampling this interval n
times.  The spectrum is calculated at the phase in the variable phase.
*/

void spectrum(I,Q,U,V,spec_info,datsw,sampling,ngrid,Cq,Cm,Br,Bt,Bp,magnetic)
float         I[],Q[],U[],V[];
spec_info_t   *spec_info;
spec_sample_t sampling;
int           datsw[],ngrid,magnetic;
float         Cq[],Cm[],Br[],Bt[],Bp[];
{
    spectra(I,Q,U,V,1,spec_info,datsw,sampling,Cq,Cm,Br,Bt,Bp,ngrid,magnetic);
}


static void doppl_spec(I,Q,U,V,spec_info,datsw,sampling,Cq,Cm,B,gamma,chi,mu,vel,magnetic)
float         I[],Q[],U[],V[];
float         Cq,Cm,B,gamma,chi,mu,vel;
int           datsw[],magnetic;
spec_info_t   spec_info;
spec_sample_t sampling;
{
    int    j,k,n,dd,nspec,os;
    float  **Sp,*tmp,chii;
    double delta,*lambda,l0,shift,temp,zeta_rt,rad_mt,tan_mt;

    if (!magnetic)  B = 0.0;

    l0      = find_line_centre(spec_info.line_id);
    zeta_rt = find_macro_turb(spec_info.line_id);
    temp   = zeta_rt*l0*TwoSqrtLog2/C_l;
    rad_mt = temp*mu; 
    tan_mt = temp*sqrt(1.0-mu*mu);

    chii = chi + spec_info.polariser;

    nspec = spec_info.nspec;
    if (!spec_info.photometric) {
        os = sampling.oversample;
        n = os*(nspec-1) + 1;
    } else {
	os = 1;
        n = nspec;
    }

    if (os == 1) {  /* No oversampling */
        Sp = pfvector(0,3);
        Sp[0] = I;  Sp[1] = Q;  Sp[2] = U;  Sp[3] = V;
    } else { /* Oversampling */
        Sp = matrix(0,3,0,n-1);
    }

    if (n > 1)
        delta = (spec_info.lamu-spec_info.laml)/(float)(n-1);
    else
	delta  = 1.0 + 1000.0*spec_info.FWHM;  /* Disables the instrumental broadening */
    shift = (1.0 - vel/C_l); /* The doppler shift multiplier for this grid cell */

    lambda = dvector(0,n-1);
#pragma _CNX no_recurrence
    for (j = 0; j < n; j++)
        lambda[j] = (j*delta + spec_info.laml)*shift;
    line_profile(spec_info.line_id,&(Sp[0][0]),&(Sp[1][0]),&(Sp[2][0]),&(Sp[3][0]),
                            n,lambda,Cq,Cm,mu,B,gamma,chii,magnetic);
    free_dvector(lambda,0,n-1);

    tmp   = vector(0,n-1);
    for (dd = 0; dd < 4; dd++) {
        if (spec_info.datsw[dd] == 1) {
	    if (!spec_info.broadened)
                smear(spec_info.FWHM,n,Sp[dd],delta);
            for (j = 0; j < n; j++) tmp[j] = Sp[dd][j];
            smear(rad_mt,n,Sp[dd],delta);
            smear(tan_mt,n,tmp,delta);
            for (j = 0; j < n; j++) Sp[dd][j] = 0.5*(Sp[dd][j] + tmp[j]);
        }
    }
    free_vector(tmp,0,n-1);

    if (os == 1) {  /* No oversampling */
        free_pfvector(Sp,0,3);
    } else { /* Oversampling */
        if (spec_info.datsw[0])
            for (j = 0, k = 0; j < nspec; j++, k+=os) I[j] = Sp[0][k];
        if (spec_info.datsw[1])
            for (j = 0, k = 0; j < nspec; j++, k+=os) Q[j] = Sp[1][k];
        if (spec_info.datsw[2])
            for (j = 0, k = 0; j < nspec; j++, k+=os) U[j] = Sp[2][k];
        if (spec_info.datsw[3])
            for (j = 0, k = 0; j < nspec; j++, k+=os) V[j] = Sp[3][k];
        free_matrix(Sp,0,3,0,n-1);
    }
}


void contin(nobs,spec_info,cont,ngrid,Cq,Cm,magnetic)
int         nobs;
spec_info_t spec_info[];
float       cont[];
int         ngrid,magnetic;
float       Cq[],Cm[];
{
    int   j,k,l;
    float C,area,phase,weight;
    cell_info_t  *cell;
    static float *dummy1=0,dummy2=0;
    static cell_deriv_t *dummy3=0;

    cell = (cell_info_t *) svector(ngrid,(int) sizeof(cell_info_t));

    for (j = 0; j < nobs; j++) {
        cont[j] = 0.0;
	for (l = 1; l <= spec_info[j].nsample; l++) {
	    phase = spec_info[j].phasec + l*(spec_info[j].phasef
	                       - spec_info[j].phasec)/(spec_info[j].nsample+1);
	    weight = 1.0 / (float) spec_info[j].nsample;
            geometry(phase,ngrid,cell,0,dummy1,dummy1,dummy1,
                                                 dummy1,dummy1,dummy1,dummy3);
            for (k = 0; k < ngrid; k++) {
                area = cell[k].parea*weight;
                if (area > 0.0) {
		    if (magnetic)
                        continuum(spec_info[j].line_id,cell[k].mu,
			    Cq[k],Cm[k],&C,&dummy2,&dummy2,magnetic);
                    else
                        continuum(spec_info[j].line_id,cell[k].mu,
			    Cq[k],dummy2,&C,&dummy2,&dummy2,magnetic);
                    cont[j] += C*area;
                }
            }
        }
    }

    free_svector(cell);
}
