/*
A program for creating simple test images, which can then
be used to generate spectral data sets with the spec program.
*/

#include <math.h>
#include <stdio.h>
#include "structures.h"
#include "constants.h"
#include "grid.h"
#include "file.h"
#include "line.h"
#include "mem_alloc.h"
#include "syn.h"
#include "utilities.h"


int main(argc,argv)
int argc;
char *argv[];
{
    int    ngrid,j,magnetic,existing,batch,Cm_used,dum1,dum2;
    char   filename[80],com[120],answer[2],*qnames[2];
    float  inc,ve,vfac,dCq,dCm,*Cq,*Cm,*Br,*Bt,*Bp,junk;

    parse_args(argc,argv,&batch,&vfac);

    init_image(&Cq,&Cm,&Br,&Bt,&Bp,&ngrid,&inc,&ve,&magnetic,&existing,batch,vfac);
    junk = 0.0; 
    image_info(&Cm_used,qnames,&dum1,&dum2,batch);

    if (!existing) {
        printf("Input the default %s : ",qnames[0]);
        scanf("%f",&dCq);
	if (batch) printf("%g\n",dCq);

	if (magnetic && Cm_used) {
            printf("Input the default %s : ",qnames[1]);
            scanf("%f",&dCm);
	    if (batch) printf("%g\n",dCm);
	} else
	    dCm = 0.0;

        for (j = 0; j < ngrid; ++j) {
            Cq[j] = dCq;
            if (magnetic) {
                Cm[j] = dCm;
		Br[j] = Bt[j] = Bp[j] = 0.0;
	    }
        }
    }

    init_grid(ngrid, inc, ve);

    printf("Do you want to initialise a multipole field (y/n) : ");
    scanf("%1s", answer);
    if ((answer[0] == 'Y') || (answer[0] == 'y')) {
	if (batch) printf("yes\n");
        dipole(Cq,Cm,Br,Bt,Bp,ngrid,magnetic,dCq,Cm_used,qnames,batch);
    } else
	if (batch) printf("no\n");

    printf("Do you want to initialise a potential field (y/n) : ");
    scanf("%1s", answer);
    if ((answer[0] == 'Y') || (answer[0] == 'y')) {
        if (batch) printf("yes\n");
        potential(Cq,Cm,Br,Bt,Bp,ngrid,magnetic,dCq,Cm_used,qnames,batch);
    } else
        if (batch) printf("no\n");

    printf("Do you want to add some spots (y/n) : ");
    scanf("%1s", answer);
    if ((answer[0] == 'Y') || (answer[0] == 'y')) {
	if (batch) printf("yes\n");
        spots(Cq,Cm,Br,Bt,Bp,ngrid,magnetic,Cm_used,qnames,batch);
    } else
	if (batch) printf("no\n");

    sprintf(com,"Synthetic Star");  
    save_bright(ngrid,Cq,Cm,Br,Bt,Bp,filename,
                              1,com,inc,ve,magnetic);
    if (batch) printf("%s\n",filename);

    free_vector(Cq,0,ngrid-1);
    if (magnetic) {
        free_vector(Cm,0,ngrid-1);
        free_vector(Br,0,ngrid-1);
        free_vector(Bt,0,ngrid-1);
        free_vector(Bp,0,ngrid-1);
    }
}


/*
Here we initialise the magnetic field and/or the abundance/brightness to
a distribution corresponding to a simple magnetic dipole.

The lattitude and longitude coordinates $(\theta,\phi)$, at the phase when
the positive pole of the dipole is at centeral meridian, are converted to
coordinates $(x,y,z)$ on the surface of a uniform sphere. The observer in
located in the $xz$ plane and the sphere is rotated around the $y$ axis so
that the $z$ axis is oriented in the direction of the dipole pole.

The angle $\theta_B$ is the inclination of the unit normal vector at point
$(x,y,z)$ to the $z$ axis. The angle $\phi_B$ is the abgle between the $xz$
plane the the plane containing the point $(x,y,z)$. The radial field component,
$B_r$, and the perpendicular component, $B_\theta$, are
given by (see, for example, Lang 1980),
$$\eqalign{
    B_r & = B_{\rm pole} \cos \theta_B \cr
    B_\theta & = (1/2) B_{\rm pole} \sin \theta_B, \cr
}$$
where $B_{\rm pole}$ if the field strength at the pole of the dipole.
Hence, the cartesian components of the field are
$$\eqalign{
    B_x & = (3/2) B_{\rm pole} \cos\theta_B \sin\theta_B \cos\phi_B \cr
    B_y & = (3/2) B_{\rm pole} \cos\theta_B \sin\theta_B \sin\phi_B \cr
    B_z & = (1/2) B_{\rm pole} (3\cos^2\theta_B - 1). \cr
}$$

The coordinate frame is then rotated so that the $z$ axis once again corresponds
to the north-south axis of the star. The cartesian field components are then
converted to radial, toroidal and poloidal fields. If desired, the image
quantity, stored in the |C_q| array, is made proportional to the magnitude
of the radial component of the field.
*/

static void dipole(Cq,Cm,Br,Bt,Bp,ngrid,magnetic,dCq,Cm_used,qnames,batch)
float Cq[],Cm[],Br[],Bt[],Bp[],dCq;
int ngrid,magnetic,Cm_used,batch;
char *qnames[];
{
    int    j,three;
    char   answer[2];
    float  dCm,dBr,dBt,dBp,mag_inc,B_pole,phase,*theta,*phi,
           A_pole,A_eqt,M_pole,M_eqt,B_quad,B_octu,B_str;
    double ci,si,theta_B,phi_B,x,x1,y,z,Bx,By,Bz,ct,st,cp,sp,
           stB,ctB,spB,cpB;

    printf("Enter the inclination of the multipole axis to the rotational axis : ");
    scanf("%f",&mag_inc);
    if (batch) printf("%g\n",mag_inc);
    mag_inc = TO_RAD(mag_inc);
    printf("Enter the phase when the positive multipole pole is at CMP : ");
    scanf("%f",&phase);
    if (batch) printf("%g\n",phase);

    if (magnetic) {
        printf("Enter the magnetic field strengths (d/q/o) at the poles : ");
        scanf("%f %f %f",&B_pole,&B_quad,&B_octu);
        if (batch) printf("%g %g %g\n",B_pole,B_quad,B_octu);
        printf("Do you want to set all three field components (y/n) : ");
        scanf("%1s", answer);
        if ((answer[0] == 'Y') || (answer[0] == 'y')) {
	    if (batch) printf("yes\n");
            three = 1;
        } else {
	    if (batch) printf("no\n");
            three = 0;
        }
    } else {
	B_pole = 1.0; B_quad = B_octu = 0.0; 
    } 

/* Arbitrary. It is needed to calculate the abundance distribution */
    B_str = sqrt(B_pole*B_pole+B_quad*B_quad+B_octu*B_octu); 
    if (B_str == 0.0) B_str = 1.0; 

    printf("Do you want to vary the %s (y/n) : ",qnames[0]);
    scanf("%1s", answer);
    if ((answer[0] == 'Y') || (answer[0] == 'y')) {
	if (batch) printf("yes\n");
        printf("Enter values for the pole and equator : ");
        scanf("%f %f",&A_pole,&A_eqt);
	if (batch) printf("%g %g\n",A_pole,A_eqt);
    } else {
	if (batch) printf("no\n");
        A_pole = A_eqt = dCq;
    }

    if (magnetic && Cm_used) {
        printf("Do you want to vary the %s (y/n) : ",qnames[1]);
        scanf("%1s", answer);
        if ((answer[0] == 'Y') || (answer[0] == 'y')) {
	    if (batch) printf("yes\n");
            printf("Enter values for the pole and equator : ");
            scanf("%f %f",&M_pole,&M_eqt);
	    if (batch) printf("%g %g\n",M_pole,M_eqt);
        } else {
	    if (batch) printf("no\n");
            M_pole = M_eqt = 0.0;
        }
    } else
	M_pole = M_eqt = 0.0;

    si = sin(mag_inc);
    ci = cos(mag_inc);
    theta = vector(0,ngrid-1);
    phi   = vector(0,ngrid-1);
    locations(phase,ngrid,theta,phi); /* Locations of the grid points */
    for (j = 0; j < ngrid; j++) { /* Loop over grid points */
        st = sin(theta[j]);  ct = cos(theta[j]);
        sp = sin(phi[j]);    cp = cos(phi[j]);
        x1 = ct*cp;
        y = ct*sp;
        z = st;
        x = x1*ci - z*si;  /* Rotation of $x$ around the $y$ axis */
        z = x1*si + z*ci;  /* Rotation of $z$ around the $y$ axis */

        if (z >= 1.0)
            theta_B = 0.0;
        else if (z <= -1.0)
            theta_B = PI;
        else
            theta_B = acos(z);
        stB = sin(theta_B);  ctB = cos(theta_B);
        if (fabs(stB) > 0.0) {
            if ((x/stB) >= 1.0)
	        phi_B = 0.0;
            else if ((x/stB) <= -1.0)
		phi_B  = PI;
            else
                phi_B = acos(x/stB);
            if (y < 0.0)
                phi_B = -phi_B;
        } else
	    phi_B = 0.0;
        spB = sin(phi_B);  cpB = cos(phi_B);
/* Dipole */ 
        Bx = 1.5*B_pole*ctB*stB*cpB;
        By = 1.5*B_pole*ctB*stB*spB;
        Bz = 0.5*B_pole*(3.0*ctB*ctB-1.0);
/* Quadrupole */ 
        Bx += 0.5*B_quad*(5.0*ctB*ctB-1.0)*stB*cpB;
        By += 0.5*B_quad*(5.0*ctB*ctB-1.0)*stB*spB;
        Bz += 0.5*B_quad*(5.0*ctB*ctB-3.0)*ctB;
/* Octopole */ 
        Bx += 0.625*B_octu*(7.0*ctB*ctB-3.0)*ctB*stB*cpB;
        By += 0.625*B_octu*(7.0*ctB*ctB-3.0)*ctB*stB*spB;
        Bz += 0.125*B_octu*(5.0*ctB*ctB*(7.0*ctB*ctB-6.0)+3.0);

        x1 = Bx*ci + Bz*si;
        Bz = -Bx*si + Bz*ci;
        Bx = x1;

        dBr = ct*cp*Bx + ct*sp*By + st*Bz;
        dCq = A_eqt + (A_pole-A_eqt)*fabs(dBr/B_str);
        dCm = M_eqt + (M_pole-M_eqt)*fabs(dBr/B_str);
        if (three) {
            dBt = -sp*Bx + cp*By;
            dBp = -cp*st*Bx - sp*st*By + ct*Bz;
        } else
            dBt = dBp = 0.0;
 
        Cq[j] = dCq;
        if (magnetic) {
            Cm[j] = dCm;
            Br[j] = dBr;
            Bt[j] = dBt;
            Bp[j] = dBp;
        }
    }
    free_vector(theta,0,ngrid-1);
    free_vector(phi,0,ngrid-1);
}


/* initialise potential field with spherical harmonics */ 

static void potential(Cq,Cm,Br,Bt,Bp,ngrid,magnetic,dCq,Cm_used,qnames,batch)
float Cq[],Cm[],Br[],Bt[],Bp[],dCq;
int ngrid,magnetic,Cm_used,batch;
char *qnames[];
{
    int k, j, l, m, n, lmax, ntot; 
    float ***L, *alm, *blm; 

    printf("l and m : "); 
    scanf("%d %d", &l, &m); 
    if (m < 0) m = -m; if (l < 0) l = -l; 
    if (m > l) m = l; 
    printf("%d %d\n",l,m); 

    lmax = l; 
    ntot = (lmax+1)*(lmax+2)-2; 
    n = l*(l+1) + 2*m - 2; 
    L = matrix3D(0,5,0,ntot-1,0,ngrid-1); 
    alm = vector(0, ntot-1); blm = vector(0, ntot-1);
    for (j = 0; j < ntot; j++) alm[j] = blm[j] = 0.0; 
/*
    alm[n] = 1000.0; blm[n] = 1000.0; 
*/
alm[2] = 300.0; blm[2] = -300.0*2.0; blm[14] = 4.0*250.0; 
printf("going to legendre\n"); 
    legendre(L,ngrid,lmax,-1,0.0); 
printf("out of legendre\n");  
    for (k = 0; k < ngrid; k++) {
        Br[k] = Bt[k] = Bp[k] = 0.0;
        for (j = 0; j < ntot; j++) {
            Br[k] += alm[j]*L[0][j][k] + blm[j]*L[3][j][k];
            Bt[k] += alm[j]*L[1][j][k] + blm[j]*L[4][j][k];
            Bp[k] += alm[j]*L[2][j][k] + blm[j]*L[5][j][k];
        }
    }
printf("here\n");  
    free_matrix3D(L,0,5,0,ntot-1,0,ngrid-1); 
    free_vector(alm,0,ntot-1); free_vector(blm,0,ntot-1); 
}
/*
Here we add circular spots to the image.
*/

static void spots(Cq,Cm,Br,Bt,Bp,ngrid,magnetic,Cm_used,qnames,batch)
float Cq[],Cm[],Br[],Bt[],Bp[];
int ngrid,magnetic,Cm_used,batch;
char *qnames[];
{
    int    j;
    char   answer[2];
    float  dCq,dCm,phase,lat,area,cosmin,*theta,*phi,
           xc,zc,cosine,dmag,dBr,dBt,dBp,fill_fac;
    double x,z;

    theta = vector(0,ngrid-1);
    phi   = vector(0,ngrid-1);
    do {
        printf("Enter the spots phase, latitude and fractional area : ");
        scanf("%f %f %f",&phase,&lat,&area);
	if (batch) printf("%g %g %g\n",phase,lat,area);
        if (!magnetic) {
            printf("Enter the %s in the spot : ",qnames[0]);
            scanf("%f",&dCq);
	    if (batch) printf("%g\n",dCq);
        } else {
	    if (Cm_used) {
                printf("Enter the spot %s, %s and field strength : ",qnames[0],qnames[1]);
                scanf("%f %f %f",&dCq,&dCm,&dmag);
	        if (batch) printf("%g %g %g\n",dCq,dCm,dmag);
	    } else {
                printf("Enter the spot %s and field strength : ",qnames[0]);
                scanf("%f %f",&dCq,&dmag);
	        if (batch) printf("%g %g\n",dCq,dmag);
		dCm = 0.0;
            }
            printf("Which field component should be set (RTPA) : ");
            scanf("%1s",answer);
	    if (answer[0] == 'T' || answer[0] == 't') {
		if (batch) printf("T\n");
                dBt = dmag;  dBr = dBp = 0.0;
	    } else if (answer[0] == 'P' || answer[0] == 'p') {
		if (batch) printf("P\n");
                dBp = dmag;  dBr = dBt = 0.0;
	    } else if (answer[0] == 'A' || answer[0] == 'a') {
		if (batch) printf("all\n");
                dBr = dBp = dBt = dmag; 
	    } else {  /* Take anything else to mean radial */
		if (batch) printf("R\n");
                dBr = dmag;  dBt = dBp = 0.0;
	    }
        }
        lat = TO_RAD(lat);
	if (area >= 0.0)
            cosmin = 1.0 - 2.0*area;
        else
	    cosmin = cos(-TO_RAD(area));
        xc = cos(lat);
        zc = sin(lat);
        locations(phase,ngrid,theta,phi);
        for (j = 0; j < ngrid; j++) {
            x = cos(theta[j])*cos(phi[j]);
            z = sin(theta[j]);
            cosine = x*xc + z*zc;
            if (cosine >= cosmin) {
                Cq[j] = dCq;
                if (magnetic && dmag) {
                    Cm[j] = dCm;
                    Br[j] = dBr;
                    Bt[j] = dBt;
                    Bp[j] = dBp;
                }
            }
        }
        printf("Do you want to add another spot (y/n) : ");
        scanf("%1s", answer);
	if (batch) printf("%s\n",answer);
    } while((answer[0] == 'Y') || (answer[0] == 'y'));
    free_vector(theta,0,ngrid-1);
    free_vector(phi,0,ngrid-1);
}
