/* This file contains a routine to calculate the Voigt and Faraday-Voigt
   functions.  The Voigt function is defined as
   $$H(a,\nu) = {a \over \pi} \int_{-\infty}^\infty {e^{-y^2} \over (\nu - y)^2 + a^2}dy$$
   and the Faraday-Voigt function is defined as
   $$F(a,\nu) = {1 \over 2\pi} \int_{-\infty}^\infty {(\nu-y)e^{-y^2} \over (\nu - y)^2 + a^2}dy.$$

   This routine is used by the default spectral modelling routines.  A lot of
   the programs time is spent in this routine so any improvements made to it
   could have a large effect on the programs overall run time.
*/

#define COMPILING_VOIGT
#include <math.h>
#include "voigt.h"


/*
The routine |fvoigt()| is a \Cee\ implementation of the \mc{FORTRAN}
routine given in the paper by Huml\'{\i}\v{c}ek$^1$.  He gives an efficient
rational function approximation for evaluating the complex probability
function
$$w(z) = e^{-z^2}\left(1 + {2i\over \sqrt{\pi}} \int_0^ze^{t^2}\,dt\right)$$
in the upper half plane.  The maximum relative error of both the real and
imaginary parts is less than $10^{-4}$.
If we put $z = \nu + ia$ then the Voigt function is given by $H = \Re{(w)}$ and
the Faraday-Voigt function is given by $F = {1 \over 2} \Im{(w)}$.

All the complex
arithmetic has been split into real and imaginary parts and coded as efficiently
as possible (I hope).  As mentioned above, improvements made here could greatly
reduce the total program running time.

{\par{\hangindent=20pt \hangafter=1
$^1$ Huml\'{\i}\v{c}ek, J. (1982) Optimized Computation of the Voigt and
                                  Complex Probability Functions.
{\it J. Quant. Spectrosc. Radiat. Transfer,} {\bf 27}, 437--444.
 \par}}
*/

void fvoigt(n,a,v,h,fv)
int n;              /* Number of points */
double a, v[];     /* Anomalous broadening coefficient, $a$, and frequency $\nu$ */
double h[], fv[];    /* Returned values of $H$ and $F$ */
{
    int      j;
    double   s,absv,tmp;
    double   w4r,w4i,nr,ni,dr,di,ur,ui,pr,pi,qr,qi;
    void     Cdivr();

/*  #pragma _CNX no_recurrence  */
    for (j = 0; j < n; j++) {
	absv = fabs(v[j]);
        s = absv + a;
        if (s >= 15.0) {
            nr = 0.5641896*a;             ni = -0.5641896*v[j];
            dr = 0.5+a*a-v[j]*v[j];       di = -2.0*a*v[j];
            tmp = 1.0/(dr*dr+di*di);
            w4r = (nr*dr+ni*di)*tmp;      w4i = (ni*dr-nr*di)*tmp;
        } else if (s >= 5.5) {
            ur = a*a-v[j]*v[j];           ui = -2.0*a*v[j];
            nr = 1.410474*a+0.5641896*(a*ur+v[j]*ui);  ni = 0.5641896*(a*ui-v[j]*ur)-1.410474*v[j];
            dr = 0.75+ur*(3.0+ur)-ui*ui;  di = 3.0*ui+2.0*ur*ui;
            tmp = 1.0/(dr*dr+di*di);
            w4r = (nr*dr+ni*di)*tmp;      w4i = (ni*dr-nr*di)*tmp;
        } else if (a >= (0.195*absv-0.176)) {
            pr = 3.778987+0.5642236*a;    pi = -0.5642236*v[j];
            qr = 11.96482+a*pr+v[j]*pi;   qi = a*pi-v[j]*pr;
            pr = 20.20933+a*qr+v[j]*qi;   pi = a*qi-v[j]*qr;
            nr = 16.4955+a*pr+v[j]*pi;    ni = a*pi-v[j]*pr;
            pr = 6.699398+a;
            qr = 21.69274+a*pr-v[j]*v[j]; qi = -v[j]*(a+pr);
            pr = 39.27121+a*qr+v[j]*qi;   pi = a*qi-v[j]*qr;
            qr = 38.82363+a*pr+v[j]*pi;   qi = a*pi-v[j]*pr;
            dr = 16.4955+a*qr+v[j]*qi;    di = a*qi-v[j]*qr;
            tmp = 1.0/(dr*dr+di*di);
            w4r = (nr*dr+ni*di)*tmp;      w4i = (ni*dr-nr*di)*tmp;
        } else {
            ur = a*a-v[j]*v[j];         ui = -2.0*a*v[j];
            pr = 1.320522-0.56419*ur;   pi = -0.56419*ui;
            qr = 35.76683-ur*pr+ui*pi;  qi = -ur*pi-ui*pr;
            pr = 219.0313-ur*qr+ui*qi;  pi = -ur*qi-ui*qr;
            qr = 1540.787-ur*pr+ui*pi;  qi = -ur*pi-ui*pr;
            pr = 3321.9905-ur*qr+ui*qi; pi = -ur*qi-ui*qr;
            qr = 36183.31-ur*pr+ui*pi;  qi = -ur*pi-ui*pr;
            nr = a*qr+v[j]*qi;          ni = a*qi-v[j]*qr;
            pr = 1.841439-ur;
            qr = 61.57037-ur*pr-ui*ui;  qi = ui*(ur-pr);
            pr = 364.2191-ur*qr+ui*qi;  pi = -ur*qi-ui*qr;
            qr = 2186.181-ur*pr+ui*pi;  qi = -ur*pi-ui*pr;
            pr = 9022.228-ur*qr+ui*qi;  pi = -ur*qi-ui*qr;
            qr = 24322.84-ur*pr+ui*pi;  qi = -ur*pi-ui*pr;
            dr = 32066.6-ur*qr+ui*qi;   di = -ur*qi-ui*qr;
            tmp = 1.0/(dr*dr+di*di);
            pr = (nr*dr+ni*di)*tmp;     pi = (ni*dr-nr*di)*tmp;
            qr = exp(ur);
            w4r = qr*cos(ui)-pr;        w4i = qr*sin(ui)-pi;
        }
        h[j]  = w4r;
        fv[j] = 0.5*w4i;
    }
}
